# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines class that describes C++ typedef declaration
"""

from . import declaration
from . import byte_info


class typedef_t(declaration.declaration_t, byte_info.byte_info):

    """describes C++ typedef declaration"""

    def __init__(self, name='', decl_type=None):
        """creates class that describes C++ typedef"""
        declaration.declaration_t.__init__(self, name)
        byte_info.byte_info.__init__(self)
        self._decl_type = decl_type
        if not isinstance(decl_type, str) and decl_type is not None:
            self.byte_size = decl_type.byte_size
            self.byte_align = decl_type.byte_align

    def _get__cmp__items(self):
        """implementation details"""
        return [self.decl_type]

    def __eq__(self, other):
        if not declaration.declaration_t.__eq__(self, other):
            return False
        return self.decl_type == other.decl_type

    __hash__ = declaration.declaration_t.__hash__

    @property
    def decl_type(self):
        """reference to the original :class:`decl_type <type_t>`"""
        return self._decl_type

    @decl_type.setter
    def decl_type(self, decl_type):
        self._decl_type = decl_type
        self.byte_size = decl_type.byte_size
        self.byte_align = decl_type.byte_align

    def i_depend_on_them(self, recursive=True):
        self._warn_deprecated()
        # Deprecated method. The cyclic import will be removed with the method
        # in the next release, so we can disable the cyclic import check here.
        from . import dependencies  # pylint: disable=R0401
        return [dependencies.dependency_info_t(self, self.decl_type)]



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from . import calldef
from . import calldef_members
from . import calldef_types
from . import declaration_utils
from . import cpptypes
from . import type_traits
from . import type_traits_classes


class free_calldef_t(calldef.calldef_t):

    """base class for "callable" declarations that defined within
    C++ namespace"""

    def __init__(self, *args, **keywords):
        calldef.calldef_t.__init__(self, *args, **keywords)

    def __str__(self):
        # Get the full name of the calldef...
        name = declaration_utils.full_name(self)
        if name[:2] == "::":
            name = name[2:]
        # Add the arguments...
        args = [str(a) for a in self.arguments]
        res = "%s(%s)" % (name, ", ".join(args))
        # Add the return type...
        if self.return_type is not None:
            res = "%s %s" % (self.return_type, res)
        # extern?
        if self.has_extern:
            res = "extern " + res
        # Append the declaration class
        cls = self.__class__.__name__
        if cls[-2:] == "_t":
            cls = cls[:-2]
        cls = cls.replace('_', ' ')
        return "%s [%s]" % (res, cls)

    def _get__cmp__call_items(self):
        """implementation details"""
        return []

    def function_type(self):
        """returns function type. See :class:`type_t` hierarchy"""
        return cpptypes.free_function_type_t(
            return_type=self.return_type,
            arguments_types=[
                arg.decl_type for arg in self.arguments])

    def create_decl_string(self, with_defaults=True):
        f_type = self.function_type()
        if with_defaults:
            return f_type.decl_string

        return f_type.partial_decl_string

    def guess_calling_convention(self):
        """This function should be overriden in the derived classes and return
        more-or-less successfull guess about calling convention"""
        return calldef_types.CALLING_CONVENTION_TYPES.UNKNOWN


class free_function_t(free_calldef_t):

    """describes free function declaration"""

    def __init__(self, *args, **keywords):
        free_calldef_t.__init__(self, *args, **keywords)

    def get_mangled_name(self):
        if not self._mangled \
           and '<' not in self.name and not self.overloads:
            # it is possible we deal with C function, so lets put it name as
            # mangled one
            return self.name

        return self._mangled


class free_operator_t(free_calldef_t, calldef_members.operator_t):

    """describes free operator declaration"""

    def __init__(self, *args, **keywords):
        free_calldef_t.__init__(self, *args, **keywords)
        calldef_members.operator_t.__init__(self, *args, **keywords)
        self.__class_types = None

    @property
    def class_types(self):
        """list of class/class declaration types, extracted from the
        operator arguments"""

        if None is self.__class_types:
            self.__class_types = []
            for type_ in self.argument_types:
                decl = None
                type_ = type_traits.remove_reference(type_)
                if type_traits_classes.is_class(type_):
                    decl = type_traits_classes.class_traits.get_declaration(
                        type_)
                elif type_traits_classes.is_class_declaration(type_):
                    tt = type_traits_classes.class_declaration_traits
                    decl = tt.get_declaration(type_)
                else:
                    pass
                if decl:
                    self.__class_types.append(decl)
        return self.__class_types



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines class that describes C++ global and member variable declaration
"""

from . import declaration
from . import class_declaration


class variable_t(declaration.declaration_t):

    """describes C++ global and member variable declaration"""

    def __init__(
            self,
            name='',
            decl_type=None,
            type_qualifiers=None,
            value=None,
            bits=None,
            mangled=None):
        """creates class that describes C++ global or member variable"""
        declaration.declaration_t.__init__(self, name)
        self._decl_type = decl_type
        self._type_qualifiers = type_qualifiers
        self._value = value
        self._bits = bits
        self._byte_offset = 0
        self._mangled = mangled

    def _get__cmp__items(self):
        """implementation details"""
        return [self.decl_type, self.type_qualifiers, self.value]

    def __eq__(self, other):
        """implementation details"""
        if not declaration.declaration_t.__eq__(self, other):
            return False
        return self.decl_type == other.decl_type \
            and self.type_qualifiers == other.type_qualifiers \
            and self.value == other.value \
            and self.bits == other.bits

    __hash__ = declaration.declaration_t.__hash__

    @property
    def decl_type(self):
        """reference to the variable :class:`decl_type <type_t>`"""
        return self._decl_type

    @decl_type.setter
    def decl_type(self, decl_type):
        self._decl_type = decl_type

    @property
    def type_qualifiers(self):
        """reference to the :class:`type_qualifiers_t` instance"""
        return self._type_qualifiers

    @type_qualifiers.setter
    def type_qualifiers(self, type_qualifiers):
        self._type_qualifiers = type_qualifiers

    @property
    def value(self):
        """string, that contains the variable value"""
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def bits(self):
        """integer, that contains information about how many bit takes
            bit field"""
        return self._bits

    @bits.setter
    def bits(self, bits):
        self._bits = bits

    @property
    def byte_offset(self):
        """integer, offset of the field from the beginning of class."""
        return self._byte_offset

    @byte_offset.setter
    def byte_offset(self, byte_offset):
        self._byte_offset = byte_offset

    @property
    def access_type(self):
        if not isinstance(self.parent, class_declaration.class_t):
            raise RuntimeError((
                "access_type functionality only available on member" +
                "variables and not on global variables"))
        return self.parent.find_out_member_access_type(self)

    @property
    def mangled(self):
        """
        Unique declaration name generated by the compiler.

        :return: the mangled name
        :rtype: str

        """

        return self.get_mangled_name()

    @mangled.setter
    def mangled(self, mangled):
        self._mangled = mangled

    def i_depend_on_them(self, recursive=True):
        # Deprecated method. The cyclic import will be removed with the method
        # in the next release, so we can disable the cyclic import check here.
        from . import dependencies  # pylint: disable=R0401
        self._warn_deprecated()
        return [dependencies.dependency_info_t(self, self.decl_type)]

    def get_mangled_name(self):
        if not self._mangled \
           and not isinstance(self.parent, class_declaration.class_t):
            return self.name

        return self._mangled



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""Logger classes and a few convenience methods."""

import os
import sys
import logging
import tempfile
import shutil
import warnings


def find_xml_generator(name="castxml", search_path=None):
    """
    Try to find a c++ parser (xml generator)

    Args:
        name (str): name of the c++ parser (e.g. castxml)
        search_path (str): helps finding castxml
                           (for example in jupyter notebooks, use sys.path)

    Returns:
        path (str), name (str): path to the xml generator and it's name


    If no c++ parser is found the function raises an exception.
    pygccxml does currently only support castxml as c++ parser.

    """

    path = shutil.which(name, path=search_path)
    if path == "" or path is None:
        raise Exception("No c++ parser found. Please install castxml.")
    return path.rstrip(), name


def _create_logger_(name):
    """Implementation detail, creates a logger."""
    logger = logging.getLogger(name)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('%(levelname)s %(message)s'))
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger


class loggers(object):
    """Class-namespace, defines a few loggers classes, used in the project."""

    cxx_parser = _create_logger_('pygccxml.cxx_parser')
    """
    Logger for C++ parser functionality

    If you set this logger level to DEBUG, you will be able to see the exact
    command line, used to invoke GCC-XML  and errors that occures during XML
    parsing

    """

    pdb_reader = _create_logger_('pygccxml.pdb_reader')
    """
    Logger for MS .pdb file reader functionality

    """

    queries_engine = _create_logger_('pygccxml.queries_engine')
    """
    Logger for query engine functionality.

    If you set this logger level to DEBUG, you will be able to see what queries
    you do against declarations tree, measure performance and may be even to
    improve it.
    Query engine reports queries and whether they are optimized or not.

    """

    declarations_cache = _create_logger_('pygccxml.declarations_cache')
    """
    Logger for declarations tree cache functionality

    If you set this logger level to DEBUG, you will be able to see what is
    exactly happens, when you read the declarations from cache file. You will
    be able to decide, whether it worse for you to use this or that cache
    strategy.

    """

    root = logging.getLogger('pygccxml')
    """
    Root logger exists for your convenience only.

    """

    all_loggers = [
        root, cxx_parser, queries_engine, declarations_cache, pdb_reader]
    """
    Contains all logger classes, defined by the class.

    """

    @staticmethod
    def set_level(level):
        """Set the same logging level for all the loggers at once."""
        for logger in loggers.all_loggers:
            logger.setLevel(level)


def remove_file_no_raise(file_name, config):
    """Removes file from disk if exception is raised."""
    # The removal can be disabled by the config for debugging purposes.
    if config.keep_xml:
        return True

    try:
        if os.path.exists(file_name):
            os.remove(file_name)
    except IOError as error:
        loggers.root.error(
            "Error occurred while removing temporary created file('%s'): %s",
            file_name, str(error))


def create_temp_file_name(suffix, prefix=None, directory=None):
    """
    Small convenience function that creates temporary files.

    This function is a wrapper around the Python built-in
    function tempfile.mkstemp.

    """

    if not prefix:
        prefix = tempfile.gettempprefix()
    fd, name = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=directory)
    file_obj = os.fdopen(fd)
    file_obj.close()
    return name


def normalize_path(some_path):
    """Return os.path.normpath(os.path.normcase(some_path))."""
    return os.path.normpath(os.path.normcase(some_path))


def contains_parent_dir(fpath, dirs):
    """
    Returns true if paths in dirs start with fpath.

    Precondition: dirs and fpath should be normalized before calling
    this function.

    """
    # Note: this function is used nowhere in pygccxml but is used
    # at least by pypluplus; so it should stay here.

    return bool([x for x in dirs if _f(fpath, x)])


def _f(fpath, dir_):
    """Helper function for contains_parent_dir function."""
    return fpath.startswith(dir_)


def get_architecture():
    """
    Returns computer architecture: 32 or 64.

    The guess is based on maxint.

    """
    if sys.maxsize == 2147483647:
        return 32
    elif sys.maxsize == 9223372036854775807:
        return 64
    else:
        raise RuntimeError("Unknown architecture")


class cached(property):
    """Convert a method into a cached attribute."""

    # The following code is cut-and-paste from this post:
    # http://groups.google.com/group/comp.lang.python/browse_thread/
    # thread/5b71896c06bd0f76/
    # Thanks to Michele Simionato

    def __init__(self, method):
        private = '_' + method.__name__

        def fget(s):
            try:
                return getattr(s, private)
            except AttributeError:
                value = method(s)
                setattr(s, private, value)
                return value

        def fdel(s):
            del s.__dict__[private]
        super(cached, self).__init__(fget, fdel=fdel)

    def reset(self):
        cls = self.__class__
        for name in dir(cls):
            attr = getattr(cls, name)
            if isinstance(attr, cached):
                delattr(self, name)


def get_tr1(name):
    """In libstd++ the tr1 namespace needs special care.

    Return either an empty string or tr1::, useful for
    appending to search patterns.

    Args:
        name (str): the name of the declaration

    Returns:
        str: an empty string or "tr1::"
    """
    tr1 = ""
    if "tr1" in name:
        tr1 = "tr1::"
    return tr1


class cxx_standard(object):
    """Helper class for parsing the C++ standard version.

    This class holds the C++ standard version the XML generator has been
    configured with, and provides helpers functions for querying C++ standard
    version related information.
    """

    __STD_CXX = {
        '-std=c++98': 199711,
        '-std=gnu++98': 199711,
        '-std=c++03': 199711,
        '-std=gnu++03': 199711,
        '-std=c++0x': 201103,
        '-std=gnu++0x': 201103,
        '-std=c++11': 201103,
        '-std=gnu++11': 201103,
        '-std=c++1y': 201402,
        '-std=gnu++1y': 201402,
        '-std=c++14': 201402,
        '-std=gnu++14': 201402,
        '-std=c++1z': 201703,
        '-std=c++17': 201703,
        '-std=gnu++1z': 201703,
        '-std=gnu++17': 201703,
        '-std=c++2a': 202002,
        '-std=gnu++2a': 202002,
        '-std=c++20': 202002,
        '-std=gnu++20': 202002,
        '-std=c++23': float('inf'),
        '-std=gnu++23': float('inf'),
    }

    def __init__(self, cflags):
        """Class constructor that parses the XML generator's command line

        Args:
            cflags (str): cflags command line arguments passed to the XML
                generator
        """
        super(cxx_standard, self).__init__()

        self._stdcxx = None
        self._is_implicit = False
        for key in cxx_standard.__STD_CXX:
            if key in cflags:
                self._stdcxx = key
                self._cplusplus = cxx_standard.__STD_CXX[key]

        if not self._stdcxx:
            if '-std=' in cflags:
                raise RuntimeError('Unknown -std=c++xx flag used')

            # Assume c++03 by default
            self._stdcxx = '-std=c++03'
            self._cplusplus = cxx_standard.__STD_CXX['-std=c++03']
            self._is_implicit = True

    @property
    def stdcxx(self):
        """Returns the -std=c++xx option passed to the constructor"""
        return self._stdcxx

    @property
    def is_implicit(self):
        """Indicates whether a -std=c++xx was specified"""
        return self._is_implicit

    @property
    def is_cxx03(self):
        """Returns true if -std=c++03 is being used"""
        return self._cplusplus == cxx_standard.__STD_CXX['-std=c++03']

    @property
    def is_cxx11(self):
        """Returns true if -std=c++11 is being used"""
        return self._cplusplus == cxx_standard.__STD_CXX['-std=c++11']

    @property
    def is_cxx11_or_greater(self):
        """Returns true if -std=c++11 or a newer standard is being used"""
        return self._cplusplus >= cxx_standard.__STD_CXX['-std=c++11']

    @property
    def is_cxx14(self):
        """Returns true if -std=c++14 is being used"""
        return self._cplusplus == cxx_standard.__STD_CXX['-std=c++14']

    @property
    def is_cxx14_or_greater(self):
        """Returns true if -std=c++14 or a newer standard is being used"""
        return self._cplusplus >= cxx_standard.__STD_CXX['-std=c++14']

    @property
    def is_cxx1z(self):
        """Returns true if -std=c++1z is being used"""
        return self._cplusplus == cxx_standard.__STD_CXX['-std=c++1z']


class DeprecationWrapper(object):
    """
    A small wrapper class useful when deprecation classes.

    This class is not part of the public API.

    """
    def __init__(self, new_target, old_name, new_name, version):
        self.new_target = new_target
        self.old_name = old_name
        self.new_name = new_name
        self.version = version

    def _warn(self):
        warnings.warn(
            self.old_name + " is deprecated. Please use " + self.new_name +
            " instead. This will be removed in version " + self.version,
            DeprecationWarning)

    def __call__(self, *args, **kwargs):
        self._warn()
        return self.new_target(*args, **kwargs)

    def __getattr__(self, attr):
        self._warn()
        return getattr(self.new_target, attr)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os
import timeit

import pygccxml.declarations

from . import source_reader
from . import declarations_cache
from . import declarations_joiner
from .. import utils


class COMPILATION_MODE(object):
    ALL_AT_ONCE = 'all at once'
    FILE_BY_FILE = 'file by file'


class file_configuration_t(object):

    """
    source code location configuration.

    The class instance uses "variant" interface to represent the following
    data:

    1) path to a C++ source file

    2) path to GCC-XML generated XML file

    3) path to a C++ source file and path to GCC-XML generated file

        In this case, if XML file does not exists, it will be created. Next
        time you will ask to parse the source file, the XML file will be used
        instead.

        Small tip: you can setup your makefile to delete XML files every time,
        the relevant source file was changed.

    4) Python string, that contains valid C++ code


    There are few functions, that will help you to construct
    :class:`file_configuration_t` object:

    * :func:`create_source_fc`

    * :func:`create_gccxml_fc`

    * :func:`create_cached_source_fc`

    * :func:`create_text_fc`

    """

    class CONTENT_TYPE(object):
        STANDARD_SOURCE_FILE = 'standard source file'
        CACHED_SOURCE_FILE = 'cached source file'
        GCCXML_GENERATED_FILE = 'gccxml generated file'
        TEXT = 'text'

    def __init__(
            self,
            data,
            start_with_declarations=None,
            content_type=CONTENT_TYPE.STANDARD_SOURCE_FILE,
            cached_source_file=None):
        object.__init__(self)
        self.__data = data
        if not start_with_declarations:
            start_with_declarations = []
        self.__start_with_declarations = start_with_declarations
        self.__content_type = content_type
        self.__cached_source_file = cached_source_file
        if not self.__cached_source_file \
           and self.__content_type == self.CONTENT_TYPE.CACHED_SOURCE_FILE:
            self.__cached_source_file = self.__data + '.xml'

    @property
    def data(self):
        return self.__data

    @property
    def start_with_declarations(self):
        return self.__start_with_declarations

    @property
    def content_type(self):
        return self.__content_type

    @property
    def cached_source_file(self):
        return self.__cached_source_file


def create_text_fc(text):
    """
    Creates :class:`parser.file_configuration_t` instance, configured to
    contain Python string, that contains valid C++ code

    :param text: C++ code
    :type text: str

    :rtype: :class:`parser.file_configuration_t`
    """

    return file_configuration_t(
        data=text,
        content_type=file_configuration_t.CONTENT_TYPE.TEXT)


def create_source_fc(header):
    """
    Creates :class:`parser.file_configuration_t` instance, configured to
    contain path to C++ source file

    :param header: path to C++ source file
    :type header: str

    :rtype: :class:`parser.file_configuration_t`
    """

    return file_configuration_t(
        data=header,
        content_type=file_configuration_t.CONTENT_TYPE.STANDARD_SOURCE_FILE)


def create_gccxml_fc(xml_file):
    """
    Creates :class:`parser.file_configuration_t` instance, configured to
    contain path to GCC-XML generated XML file.

    :param xml_file: path to GCC-XML generated XML file
    :type xml_file: str

    :rtype: :class:`parser.file_configuration_t`
    """

    return file_configuration_t(
        data=xml_file,
        content_type=file_configuration_t.CONTENT_TYPE.GCCXML_GENERATED_FILE)


def create_cached_source_fc(header, cached_source_file):
    """
    Creates :class:`parser.file_configuration_t` instance, configured to
    contain path to GCC-XML generated XML file and C++ source file. If XML file
    does not exists, it will be created and used for parsing. If XML file
    exists, it will be used for parsing.

    :param header: path to C++ source file
    :type header: str

    :param cached_source_file: path to GCC-XML generated XML file
    :type cached_source_file: str

    :rtype: :class:`parser.file_configuration_t`
    """

    return file_configuration_t(
        data=header,
        cached_source_file=cached_source_file,
        content_type=file_configuration_t.CONTENT_TYPE.CACHED_SOURCE_FILE)


class project_reader_t(object):

    """parses header files and returns the contained declarations"""

    def __init__(self, config, cache=None, decl_factory=None):
        """
        :param config: GCCXML configuration
        :type config: :class:xml_generator_configuration_t

        :param cache: declaration cache, by default a cache functionality will
                      not be used
        :type cache: :class:`cache_base_t` instance or `str`

        :param decl_factory: declaration factory
        :type decl_factory: :class:`decl_factory_t`
        """

        self.__config = config
        self.__dcache = None
        if isinstance(cache, declarations_cache.cache_base_t):
            self.__dcache = cache
        elif isinstance(cache, str):
            self.__dcache = declarations_cache.file_cache_t(cache)
        else:
            self.__dcache = declarations_cache.dummy_cache_t()
        self.__decl_factory = decl_factory
        if not decl_factory:
            self.__decl_factory = pygccxml.declarations.decl_factory_t()

        self.logger = utils.loggers.cxx_parser
        self.__xml_generator_from_xml_file = None

    @property
    def xml_generator_from_xml_file(self):
        """
        Configuration object containing information about the xml generator
        read from the xml file.

        Returns:
            utils.xml_generators: configuration object
        """
        return self.__xml_generator_from_xml_file

    @staticmethod
    def get_os_file_names(files):
        """
        returns file names

        :param files: list of strings and\\or :class:`file_configuration_t`
                      instances.
        :type files: list
        """

        fnames = []
        for f in files:
            if isinstance(f, str):
                fnames.append(f)
            elif isinstance(f, file_configuration_t):
                if f.content_type in (
                        file_configuration_t.CONTENT_TYPE.STANDARD_SOURCE_FILE,
                        file_configuration_t.CONTENT_TYPE.CACHED_SOURCE_FILE):

                    fnames.append(f.data)
            else:
                pass
        return fnames

    def read_files(
            self,
            files,
            compilation_mode=COMPILATION_MODE.FILE_BY_FILE):
        """
        parses a set of files

        :param files: list of strings and\\or :class:`file_configuration_t`
                      instances.
        :type files: list

        :param compilation_mode: determines whether the files are parsed
                                 individually or as one single chunk
        :type compilation_mode: :class:`COMPILATION_MODE`
        :rtype: [:class:`declaration_t`]
        """

        if compilation_mode == COMPILATION_MODE.ALL_AT_ONCE \
           and len(files) == len(self.get_os_file_names(files)):
            return self.__parse_all_at_once(files)
        else:
            if compilation_mode == COMPILATION_MODE.ALL_AT_ONCE:
                msg = ''.join([
                    "Unable to parse files using ALL_AT_ONCE mode. ",
                    "There is some file configuration that is not file. ",
                    "pygccxml.parser.project_reader_t switches to ",
                    "FILE_BY_FILE mode."])
                self.logger.warning(msg)
            return self.__parse_file_by_file(files)

    def __parse_file_by_file(self, files):
        namespaces = []
        config = self.__config.clone()
        self.logger.debug("Reading project files: file by file")
        for prj_file in files:

            if isinstance(prj_file, file_configuration_t):
                del config.start_with_declarations[:]
                config.start_with_declarations.extend(
                    prj_file.start_with_declarations)
                header = prj_file.data
                content_type = prj_file.content_type
            else:
                config = self.__config
                header = prj_file
                content_type = \
                    file_configuration_t.CONTENT_TYPE.STANDARD_SOURCE_FILE

            reader = source_reader.source_reader_t(
                config,
                self.__dcache,
                self.__decl_factory)

            if content_type == \
                    file_configuration_t.CONTENT_TYPE.STANDARD_SOURCE_FILE:
                self.logger.info('Parsing source file "%s" ... ', header)
                decls = reader.read_file(header)
            elif content_type == \
                    file_configuration_t.CONTENT_TYPE.GCCXML_GENERATED_FILE:
                self.logger.info('Parsing xml file "%s" ... ', header)
                decls = reader.read_xml_file(header)
            elif content_type == \
                    file_configuration_t.CONTENT_TYPE.CACHED_SOURCE_FILE:
                # TODO: raise error when header file does not exist
                if not os.path.exists(prj_file.cached_source_file):
                    dir_ = os.path.split(prj_file.cached_source_file)[0]
                    if dir_ and not os.path.exists(dir_):
                        os.makedirs(dir_)
                    self.logger.info(
                        'Creating xml file "%s" from source file "%s" ... ',
                        prj_file.cached_source_file, header)
                    reader.create_xml_file(header, prj_file.cached_source_file)
                self.logger.info(
                    'Parsing xml file "%s" ... ',
                    prj_file.cached_source_file)
                decls = reader.read_xml_file(prj_file.cached_source_file)
            else:
                decls = reader.read_string(header)
            self.__xml_generator_from_xml_file = \
                reader.xml_generator_from_xml_file
            namespaces.append(decls)

        self.logger.debug("Flushing cache... ")
        start_time = timeit.default_timer()
        self.__dcache.flush()
        self.logger.debug(
            "Cache has been flushed in %.1f secs",
            (timeit.default_timer() - start_time))
        answer = []
        self.logger.debug("Joining namespaces ...")
        for file_nss in namespaces:
            answer = self._join_top_namespaces(answer, file_nss)
        self.logger.debug("Joining declarations ...")
        for ns in answer:
            if isinstance(ns, pygccxml.declarations.namespace_t):
                declarations_joiner.join_declarations(ns)
        leaved_classes = self._join_class_hierarchy(answer)
        types = self.__declarated_types(answer)
        self.logger.debug("Relinking declared types ...")
        self._relink_declarated_types(leaved_classes, types)
        declarations_joiner.bind_aliases(
            pygccxml.declarations.make_flatten(answer))
        return answer

    def __parse_all_at_once(self, files):
        config = self.__config.clone()
        self.logger.debug("Reading project files: all at once")
        header_content = []
        for header in files:
            if isinstance(header, file_configuration_t):
                del config.start_with_declarations[:]
                config.start_with_declarations.extend(
                    header.start_with_declarations)
                header_content.append(
                    '#include "%s" %s' %
                    (header.data, os.linesep))
            else:
                header_content.append(
                    '#include "%s" %s' %
                    (header, os.linesep))
        declarations = self.read_string(''.join(header_content))
        declarations = self._join_top_namespaces(declarations, [])
        for ns in declarations:
            if isinstance(ns, pygccxml.declarations.namespace_t):
                declarations_joiner.join_declarations(ns)
        return declarations

    def read_string(self, content):
        """Parse a string containing C/C++ source code.

        :param content: C/C++ source code.
        :type content: str
        :rtype: Declarations
        """
        reader = source_reader.source_reader_t(
            self.__config,
            None,
            self.__decl_factory)
        decls = reader.read_string(content)
        self.__xml_generator_from_xml_file = reader.xml_generator_from_xml_file
        return decls

    def read_xml(self, file_configuration):
        """parses C++ code, defined on the file_configurations and returns
        GCCXML generated file content"""

        xml_file_path = None
        delete_xml_file = True
        fc = file_configuration
        reader = source_reader.source_reader_t(
            self.__config,
            None,
            self.__decl_factory)
        try:
            if fc.content_type == fc.CONTENT_TYPE.STANDARD_SOURCE_FILE:
                self.logger.info('Parsing source file "%s" ... ', fc.data)
                xml_file_path = reader.create_xml_file(fc.data)
            elif fc.content_type == \
                    file_configuration_t.CONTENT_TYPE.GCCXML_GENERATED_FILE:
                self.logger.info('Parsing xml file "%s" ... ', fc.data)
                xml_file_path = fc.data
                delete_xml_file = False
            elif fc.content_type == fc.CONTENT_TYPE.CACHED_SOURCE_FILE:
                # TODO: raise error when header file does not exist
                if not os.path.exists(fc.cached_source_file):
                    dir_ = os.path.split(fc.cached_source_file)[0]
                    if dir_ and not os.path.exists(dir_):
                        os.makedirs(dir_)
                    self.logger.info(
                        'Creating xml file "%s" from source file "%s" ... ',
                        fc.cached_source_file, fc.data)
                    xml_file_path = reader.create_xml_file(
                        fc.data,
                        fc.cached_source_file)
                else:
                    xml_file_path = fc.cached_source_file
            else:
                xml_file_path = reader.create_xml_file_from_string(fc.data)
            with open(xml_file_path, "r") as xml_file:
                xml = xml_file.read()
            utils.remove_file_no_raise(xml_file_path, self.__config)
            self.__xml_generator_from_xml_file = \
                reader.xml_generator_from_xml_file
            return xml
        finally:
            if xml_file_path and delete_xml_file:
                utils.remove_file_no_raise(xml_file_path, self.__config)

    @staticmethod
    def _join_top_namespaces(main_ns_list, other_ns_list):
        answer = main_ns_list[:]
        for other_ns in other_ns_list:
            same_name_namespaces = pygccxml.declarations.find_all_declarations(
                answer,
                decl_type=pygccxml.declarations.namespace_t,
                name=other_ns.name,
                parent=None,  # top-level only
                recursive=False
            )
            if len(same_name_namespaces) == 0:
                answer.append(other_ns)
            elif len(same_name_namespaces) == 1:
                same_name_namespaces[0].take_parenting(other_ns)
            else:
                primary_ns = same_name_namespaces[0]
                for extra_ns in same_name_namespaces[1:]:
                    primary_ns.take_parenting(extra_ns)
                    answer.remove(extra_ns)
                # then unify the new other_ns
                primary_ns.take_parenting(other_ns)
        return answer

    @staticmethod
    def _create_key(decl):
        return (
            decl.location.as_tuple(),
            tuple(pygccxml.declarations.declaration_path(decl)))

    def _join_class_hierarchy(self, namespaces):
        classes = [
            decl for decl in pygccxml.declarations.make_flatten(namespaces)
            if isinstance(decl, pygccxml.declarations.class_t)]
        leaved_classes = {}
        # selecting classes to leave
        for class_ in classes:
            key = self._create_key(class_)
            if key not in leaved_classes:
                leaved_classes[key] = class_
        # replacing base and derived classes with those that should be leave
        # also this loop will add missing derived classes to the base
        for class_ in classes:
            leaved_class = leaved_classes[self._create_key(class_)]
            for base_info in class_.bases:
                leaved_base = leaved_classes[
                    self._create_key(base_info.related_class)]
                # treating base class hierarchy of leaved_class
                leaved_base_info = pygccxml.declarations.hierarchy_info_t(
                    related_class=leaved_base, access=base_info.access)
                if leaved_base_info not in leaved_class.bases:
                    leaved_class.bases.append(leaved_base_info)
                else:
                    index = leaved_class.bases.index(leaved_base_info)
                    leaved_class.bases[
                        index].related_class = leaved_base_info.related_class
                # treating derived class hierarchy of leaved_base
                leaved_derived_for_base_info = \
                    pygccxml.declarations.hierarchy_info_t(
                        related_class=leaved_class,
                        access=base_info.access)
                if leaved_derived_for_base_info not in leaved_base.derived:
                    leaved_base.derived.append(leaved_derived_for_base_info)
                else:
                    index = leaved_base.derived.index(
                        leaved_derived_for_base_info)
                    leaved_base.derived[index].related_class = \
                        leaved_derived_for_base_info.related_class
            for derived_info in class_.derived:
                leaved_derived = leaved_classes[
                    self._create_key(
                        derived_info.related_class)]
                # treating derived class hierarchy of leaved_class
                leaved_derived_info = pygccxml.declarations.hierarchy_info_t(
                    related_class=leaved_derived, access=derived_info.access)
                if leaved_derived_info not in leaved_class.derived:
                    leaved_class.derived.append(leaved_derived_info)
                # treating base class hierarchy of leaved_derived
                leaved_base_for_derived_info = \
                    pygccxml.declarations.hierarchy_info_t(
                        related_class=leaved_class,
                        access=derived_info.access)
                if leaved_base_for_derived_info not in leaved_derived.bases:
                    leaved_derived.bases.append(leaved_base_for_derived_info)
        # this loops remove instance we from parent.declarations
        for class_ in classes:
            key = self._create_key(class_)
            if id(leaved_classes[key]) == id(class_):
                continue
            else:
                if class_.parent:
                    declarations = class_.parent.declarations
                else:
                    # yes, we are talking about global class that doesn't
                    # belong to any namespace. Usually is compiler generated
                    # top level classes
                    declarations = namespaces
                declarations_ids = [id(decl) for decl in declarations]
                del declarations[declarations_ids.index(id(class_))]
        return leaved_classes

    @staticmethod
    def _create_name_key(decl):
        # Not all declarations have a mangled name with castxml
        # we can only rely on the name
        if decl.mangled is not None:
            # gccxml
            return decl.location.as_tuple(), decl.mangled

        # castxml
        return decl.location.as_tuple(), decl.name

    def _relink_declarated_types(self, leaved_classes, declarated_types):

        mangled_leaved_classes = {}
        for leaved_class in leaved_classes.values():
            mangled_leaved_classes[
                self._create_name_key(leaved_class)] = leaved_class

        for decl_wrapper_type in declarated_types:
            # it is possible, that cache contains reference to dropped class
            # We need to clear it
            decl_wrapper_type.cache.reset()
            if isinstance(
                    decl_wrapper_type.declaration,
                    pygccxml.declarations.class_t):
                key = self._create_key(decl_wrapper_type.declaration)
                if key in leaved_classes:
                    decl_wrapper_type.declaration = leaved_classes[key]
                else:

                    name = decl_wrapper_type.declaration._name
                    if name == "":
                        # Happens with gcc5, castxml + std=c++11
                        # See issue #45
                        continue
                    if name.startswith("__vmi_class_type_info_pseudo"):
                        continue
                    if name == "rebind<std::__tree_node" + \
                            "<std::basic_string<char>, void *> >":
                        continue
                    if name == "type":
                        continue

                    msg = []
                    msg.append(
                        "Unable to find out actual class definition: '%s'." %
                        name)
                    msg.append((
                        "Class definition has been changed from one " +
                        "compilation to an other."))
                    msg.append((
                        "Why did it happen to me? Here is a short list " +
                        "of reasons: "))
                    msg.append((
                        "    1. There are different preprocessor " +
                        "definitions applied on same file during compilation"))
                    msg.append("    2. Bug in pygccxml.")
                    raise Exception(os.linesep.join(msg))
            elif isinstance(
                    decl_wrapper_type.declaration,
                    pygccxml.declarations.class_declaration_t):
                key = self._create_name_key(decl_wrapper_type.declaration)
                if key in mangled_leaved_classes:
                    decl_wrapper_type.declaration = mangled_leaved_classes[key]

    @staticmethod
    def __declarated_types(namespaces):
        def get_from_type(cpptype):
            if not cpptype:
                return []
            elif isinstance(cpptype, pygccxml.declarations.fundamental_t):
                return []
            elif isinstance(cpptype, pygccxml.declarations.declarated_t):
                return [cpptype]
            elif isinstance(cpptype, pygccxml.declarations.compound_t):
                return get_from_type(cpptype.base)
            elif isinstance(cpptype, pygccxml.declarations.calldef_type_t):
                types = get_from_type(cpptype.return_type)
                for arg in cpptype.arguments_types:
                    types.extend(get_from_type(arg))
                return types
            assert isinstance(
                cpptype,
                (pygccxml.declarations.unknown_t,
                 pygccxml.declarations.ellipsis_t))
            return []
        types = []
        for decl in pygccxml.declarations.make_flatten(namespaces):
            if isinstance(decl, pygccxml.declarations.calldef_t):
                types.extend(get_from_type(decl.function_type()))
            elif isinstance(
                    decl, (pygccxml.declarations.typedef_t,
                           pygccxml.declarations.variable_t)):
                types.extend(get_from_type(decl.decl_type))
        return types



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Free function invocation parser

The parser is able to extract function name and list of arguments from a
function invocation statement. For example, for the following code

.. code-block:: c++

  do_something( x1, x2, x3 )

the parser will extract
- function name - `do_something`
- argument names - `[ x1, x2, x3 ]`

"""

from . import pattern_parser

__THE_PARSER = pattern_parser.parser_t('(', ')', ',')


def is_call_invocation(declaration_string):
    """
    Returns True if `declaration_string` is a function invocation.

    :param declaration_string: string that should be checked for pattern.
    :type declaration_string: str

    :rtype: bool

    """
    return __THE_PARSER.has_pattern(declaration_string)


def name(declaration_string):
    """
    Returns the name of a function.

    :type declaration_string: str
    :rtype: str

    """
    return __THE_PARSER.name(declaration_string)


def args(declaration_string):
    """
    Returns list of function arguments

    :type declaration_string: str
    :rtype: [str]

    """
    return __THE_PARSER.args(declaration_string)


NOT_FOUND = __THE_PARSER.NOT_FOUND


def find_args(text, start=None):
    """
    Finds arguments within function invocation.

    :type text: str
    :rtype: [ arguments ] or :data:NOT_FOUND if arguments could not be found.

    """
    return __THE_PARSER.find_args(text, start)


def split(declaration_string):
    """
    Returns (name, [arguments] )

    """
    return __THE_PARSER.split(declaration_string)


def split_recursive(declaration_string):
    """
    Returns [(name, [arguments])].

    """
    return __THE_PARSER.split_recursive(declaration_string)


def join(name_, args_, arg_separator=None):
    """
    Returns name( argument_1, argument_2, ..., argument_n ).

    """
    return __THE_PARSER.join(name_, args_, arg_separator)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines default declarations factory class
"""

from .calldef_members import member_function_t
from .calldef_members import constructor_t
from .calldef_members import destructor_t
from .calldef_members import member_operator_t
from .calldef_members import casting_operator_t
from .comment import comment_t
from .free_calldef import free_function_t
from .free_calldef import free_operator_t
from .enumeration import enumeration_t
from .namespace import namespace_t
from .class_declaration import class_t
from .class_declaration import class_declaration_t
from .typedef import typedef_t
from .variable import variable_t


class decl_factory_t(object):

    """
    declarations factory class
    """

    def __init__(self):
        """creates declarations factory"""
        object.__init__(self)

    def create_member_function(self, *arguments, **keywords):
        """creates instance of class that describes member function
        declaration"""
        return member_function_t(*arguments, **keywords)

    def create_constructor(self, *arguments, **keywords):
        """creates instance of class that describes constructor declaration"""
        return constructor_t(*arguments, **keywords)

    def create_destructor(self, *arguments, **keywords):
        """creates instance of class that describes destructor declaration"""
        return destructor_t(*arguments, **keywords)

    def create_member_operator(self, *arguments, **keywords):
        """creates instance of class that describes member operator
        declaration"""
        return member_operator_t(*arguments, **keywords)

    def create_casting_operator(self, *arguments, **keywords):
        """creates instance of class that describes casting operator
        declaration"""
        return casting_operator_t(*arguments, **keywords)

    def create_free_function(self, *arguments, **keywords):
        """creates instance of class that describes free function
        declaration"""
        return free_function_t(*arguments, **keywords)

    def create_free_operator(self, *arguments, **keywords):
        """creates instance of class that describes free operator
        declaration"""
        return free_operator_t(*arguments, **keywords)

    def create_class_declaration(self, *arguments, **keywords):
        """creates instance of class that describes class declaration"""
        return class_declaration_t(*arguments, **keywords)

    def create_class(self, *arguments, **keywords):
        """creates instance of class that describes class definition
        declaration"""
        return class_t(*arguments, **keywords)

    def create_enumeration(self, *arguments, **keywords):
        """creates instance of class that describes enumeration declaration"""
        return enumeration_t(*arguments, **keywords)

    def create_namespace(self, *arguments, **keywords):
        """creates instance of class that describes namespace declaration"""
        return namespace_t(*arguments, **keywords)

    def create_typedef(self, *arguments, **keywords):
        """creates instance of class that describes typedef declaration"""
        return typedef_t(*arguments, **keywords)

    def create_variable(self, *arguments, **keywords):
        """creates instance of class that describes variable declaration"""
        return variable_t(*arguments, **keywords)

    def create_comment(self, *arguments, **keywords):
        """creates instance of class that describes variable declaration"""
        return comment_t(*arguments, **keywords)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Describe a C++ namespace declaration.

"""

from . import scopedef
from . import declaration_utils


class namespace_t(scopedef.scopedef_t):
    """
    Describes C++ namespace.

    """

    def __init__(self, name='', declarations=None):
        """
        Creates an object that describes a C++ namespace declaration.

        Args:
            name (str): name of the namespace
            declarations (list[declaration_t]): list of declarations

        """
        scopedef.scopedef_t.__init__(self, name)
        if not declarations:
            declarations = []
        # List of all declarations belonging to this namespace
        self._declarations = declarations

    def __str__(self):
        name = declaration_utils.full_name(self)
        if name != "::" and name[:2] == "::":
            name = name[2:]
        return "%s [namespace]" % name

    def _get__cmp__scope_items(self):
        """
        Implementation detail.

        """
        return [self.declarations.sort()]

    def _get_declarations_impl(self):
        return self._declarations

    @property
    def declarations(self):
        """
        List of children declarations.

        Returns:
            list[declaration_t]
        """
        return scopedef.scopedef_t.declarations.fget(self)

    @declarations.setter
    def declarations(self, declarations):
        """
        Set list of all declarations defined in the namespace.

        Args:
            declarations (list[declaration_t]): list of declarations

        """
        self._declarations = declarations

    def take_parenting(self, inst):
        """
        Takes parenting from inst and transfers it to self.

        Args:
            inst (namespace_t): a namespace declaration

        """

        if self is inst:
            return
        for decl in inst.declarations:
            decl.parent = self
            self.declarations.append(decl)
        inst.declarations = []

    def adopt_declaration(self, decl):
        self.declarations.append(decl)
        decl.parent = self
        decl.cache.reset()

    def remove_declaration(self, decl):
        """
        Removes declaration from members list.

        :param decl: declaration to be removed
        :type decl: :class:`declaration_t`

        """

        del self.declarations[self.declarations.index(decl)]
        decl.cache.reset()
        # add more comment about this.
        # if not keep_parent:
        #    decl.parent=None

    def namespace(self, name=None, function=None, recursive=None):
        """
        Returns reference to namespace declaration that matches
        a defined criteria.

        """

        return (
            self._find_single(
                scopedef.scopedef_t._impl_matchers[namespace_t.namespace],
                name=name,
                function=function,
                recursive=recursive)
        )

    def namespaces(
            self,
            name=None,
            function=None,
            recursive=None,
            allow_empty=None):
        """
        Returns a set of namespace declarations that match
        a defined criteria.

        """

        return (
            self._find_multiple(
                scopedef.scopedef_t._impl_matchers[namespace_t.namespace],
                name=name,
                function=function,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def free_function(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """
        Returns reference to free function declaration that matches
        a defined criteria.

        """

        return (
            self._find_single(
                scopedef.scopedef_t._impl_matchers[namespace_t.free_function],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[namespace_t.free_function],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def free_functions(
            self,
            name=None,
            function=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """
        Returns a set of free function declarations that match
        a defined criteria.

        """

        return (
            self._find_multiple(
                scopedef.scopedef_t._impl_matchers[namespace_t.free_function],
                name=name,
                function=function,
                decl_type=self._impl_decl_types[namespace_t.free_function],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def free_operator(
            self,
            name=None,
            function=None,
            symbol=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None):
        """
        Returns reference to free operator declaration that matches
        a defined criteria.

        """
        return (
            self._find_single(
                scopedef.scopedef_t._impl_matchers[namespace_t.free_operator],
                name=self._build_operator_name(name, function, symbol),
                symbol=symbol,
                function=self._build_operator_function(name, function),
                decl_type=self._impl_decl_types[namespace_t.free_operator],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive)
        )

    def free_operators(
            self,
            name=None,
            function=None,
            symbol=None,
            return_type=None,
            arg_types=None,
            header_dir=None,
            header_file=None,
            recursive=None,
            allow_empty=None):
        """
        Returns a set of free operator declarations that match
        a defined criteria.

        """

        return (
            self._find_multiple(
                scopedef.scopedef_t._impl_matchers[namespace_t.free_operator],
                name=self._build_operator_name(name, function, symbol),
                symbol=symbol,
                function=self._build_operator_function(name, function),
                decl_type=self._impl_decl_types[namespace_t.free_operator],
                return_type=return_type,
                arg_types=arg_types,
                header_dir=header_dir,
                header_file=header_file,
                recursive=recursive,
                allow_empty=allow_empty)
        )

    def i_depend_on_them(self, recursive=True):
        self._warn_deprecated()
        answer = []
        if recursive:
            for decl in self.declarations:
                answer.extend(decl.i_depend_on_them())
        return answer


def get_global_namespace(decls):
    """
    Get the global namespace (::) from a declaration tree.

    Args:
        decls (list[declaration_t]): a list of declarations

    Returns:
        namespace_t: the global namespace_t object (::)

    """
    found = [
        decl for decl in scopedef.make_flatten(decls) if decl.name == '::' and
        isinstance(decl, namespace_t)]
    if len(found) == 1:
        return found[0]
    raise RuntimeError("Unable to find global namespace.")



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Define a few algorithms that deal with different properties of std containers.

"""

import string
from . import cpptypes
from . import templates
from . import type_traits
from . import namespace
from . import class_declaration
from . import traits_impl_details
from .. import utils

std_namespaces = ('std', 'stdext', '__gnu_cxx')

# Take into account different equivalences (with or without spaces)
string_equivalences = type_traits.string_equivalences + \
    type_traits.normalized_string_equivalences
string_equivalences = [
    v for v in string_equivalences if not v == "std::string"]
wstring_equivalences = type_traits.wstring_equivalences + \
    type_traits.normalized_wstring_equivalences
wstring_equivalences = [
    v for v in wstring_equivalences if not v == "std::wstring"]


class defaults_eraser(object):

    def __init__(self, unordered_maps_and_sets):
        self.unordered_maps_and_sets = unordered_maps_and_sets

    def normalize(self, type_str):
        return type_str.replace(' ', '')

    @staticmethod
    def replace_basic_string(cls_name):
        strings = {
            "std::string": string_equivalences,
            "std::wstring": wstring_equivalences
            }

        new_name = cls_name
        for short_name, long_names in strings.items():
            for lname in long_names:
                new_name = new_name.replace(lname, short_name)
        return new_name

    def decorated_call_prefix(self, cls_name, text, doit):
        has_text = cls_name.startswith(text)
        if has_text:
            cls_name = cls_name[len(text):]
        answer = doit(cls_name)
        if has_text:
            answer = text + answer
        return answer

    def decorated_call_suffix(self, cls_name, text, doit):
        has_text = cls_name.endswith(text)
        if has_text:
            cls_name = cls_name[: len(text)]
        answer = doit(cls_name)
        if has_text:
            answer = answer + text
        return answer

    def erase_call(self, cls_name):
        c_traits = find_container_traits(cls_name)
        if not c_traits:
            return cls_name
        return c_traits.remove_defaults(cls_name)

    def no_std(self, cls_name):
        return self.decorated_call_prefix(
            cls_name, 'std::' + utils.get_tr1(cls_name), self.erase_call)

    def no_stdext(self, cls_name):
        return self.decorated_call_prefix(
            cls_name, 'stdext::', self.no_std)

    def no_gnustd(self, cls_name):
        return self.decorated_call_prefix(
            cls_name, '__gnu_cxx::', self.no_stdext)

    def no_const(self, cls_name):
        return self.decorated_call_prefix(
            cls_name, 'const ', self.no_gnustd)

    def no_end_const(self, cls_name):
        return self.decorated_call_suffix(
            cls_name, ' const', self.no_const)

    def erase_recursive(self, cls_name):
        return self.no_end_const(cls_name)

    def erase_allocator(self, cls_name, default_allocator='std::allocator'):
        c_name, c_args = templates.split(cls_name)
        if len(c_args) != 2:
            return
        value_type = c_args[0]
        tmpl = string.Template(
            "$container<$value_type, $allocator<$value_type>>")
        tmpl = tmpl.substitute(
            container=c_name,
            value_type=value_type,
            allocator=default_allocator)
        if self.normalize(cls_name) == \
                self.normalize(tmpl):
            return templates.join(
                c_name, [self.erase_recursive(value_type)])

    def erase_container(self, cls_name, default_container_name='std::deque'):
        c_name, c_args = templates.split(cls_name)
        if len(c_args) != 2:
            return
        value_type = c_args[0]
        dc_no_defaults = self.erase_recursive(c_args[1])
        if self.normalize(dc_no_defaults) == self.normalize(
                templates.join(default_container_name, [value_type])):
            return templates.join(
                c_name, [self.erase_recursive(value_type)])

    def erase_container_compare(
            self,
            cls_name,
            default_container_name='std::vector',
            default_compare='std::less'):
        c_name, c_args = templates.split(cls_name)
        if len(c_args) != 3:
            return
        dc_no_defaults = self.erase_recursive(c_args[1])
        if self.normalize(dc_no_defaults) != self.normalize(
                templates.join(default_container_name, [c_args[0]])):
            return
        dcomp_no_defaults = self.erase_recursive(c_args[2])
        if self.normalize(dcomp_no_defaults) != self.normalize(
                templates.join(default_compare, [c_args[0]])):
            return
        value_type = self.erase_recursive(c_args[0])
        return templates.join(c_name, [value_type])

    def erase_compare_allocator(
            self,
            cls_name,
            default_compare='std::less',
            default_allocator='std::allocator'):
        c_name, c_args = templates.split(cls_name)
        if len(c_args) != 3:
            return
        value_type = c_args[0]
        tmpl = string.Template(
            "$container<$value_type, $compare<$value_type>, " +
            "$allocator<$value_type>>")
        tmpl = tmpl.substitute(
            container=c_name,
            value_type=value_type,
            compare=default_compare,
            allocator=default_allocator)
        if self.normalize(cls_name) == \
                self.normalize(tmpl):
            return templates.join(
                c_name, [self.erase_recursive(value_type)])

    def erase_map_compare_allocator(
            self,
            cls_name,
            default_compare='std::less',
            default_allocator='std::allocator'):
        c_name, c_args = templates.split(cls_name)
        if len(c_args) != 4:
            return
        key_type = c_args[0]
        mapped_type = c_args[1]
        tmpls = [
            string.Template(
                "$container<$key_type, $mapped_type, $compare<$key_type>, " +
                "$allocator<std::pair<const $key_type, $mapped_type>>>"),
            string.Template(
                "$container<$key_type, $mapped_type, $compare<$key_type>, " +
                "$allocator<std::pair< $key_type const, $mapped_type>>>"),
            string.Template(
                "$container<$key_type, $mapped_type, $compare<$key_type>, " +
                "$allocator<std::pair<$key_type, $mapped_type>>>")]
        for tmpl in tmpls:
            tmpl = tmpl.substitute(
                container=c_name,
                key_type=key_type,
                mapped_type=mapped_type,
                compare=default_compare,
                allocator=default_allocator)
            if self.normalize(cls_name) == self.normalize(tmpl):
                return templates.join(
                    c_name,
                    [self.erase_recursive(key_type),
                     self.erase_recursive(mapped_type)])

    def erase_hash_allocator(self, cls_name):
        c_name, c_args = templates.split(cls_name)
        if len(c_args) < 3:
            return

        default_less = 'std::less'
        default_equal_to = 'std::equal_to'
        default_allocator = 'std::allocator'

        if len(c_args) == 3:
            default_hash = 'hash_compare'
            tmpl = (
                "$container<$value_type, $hash<$value_type, " +
                "$less<$value_type>>, $allocator<$value_type>>")
        elif len(c_args) == 4:
            default_hash = 'hash'
            tmpl = (
                "$container<$value_type, $hash<$value_type>, " +
                "$equal_to<$value_type>, $allocator<$value_type>>")
        else:
            return

        value_type = c_args[0]
        template = string.Template(tmpl)
        for ns in std_namespaces:
            inst = template.substitute(
                container=c_name,
                value_type=value_type,
                hash=ns + '::' + utils.get_tr1(cls_name) + default_hash,
                less=default_less,
                equal_to=default_equal_to,
                allocator=default_allocator)
            if self.normalize(cls_name) == \
                    self.normalize(inst):
                return templates.join(
                    c_name, [self.erase_recursive(value_type)])

    def erase_hashmap_compare_allocator(self, cls_name):
        c_name, c_args = templates.split(cls_name)

        if self.unordered_maps_and_sets:
            default_less_or_hash = 'std::hash'
        else:
            default_less_or_hash = 'std::less'
        default_allocator = 'std::allocator'
        default_equal_to = 'std::equal_to'

        if len(c_args) > 2:
            key_type = c_args[0]
            mapped_type = c_args[1]
        else:
            return

        if len(c_args) == 4:
            default_hash = 'hash_compare'
            tmpl = string.Template(
                "$container<$key_type, $mapped_type, " +
                "$hash<$key_type, $less<$key_type>>, " +
                "$allocator<std::pair<const $key_type, $mapped_type>>>")
            if key_type.startswith('const ') or key_type.endswith(' const'):
                tmpl = string.Template(
                    "$container<$key_type, $mapped_type, $hash<$key_type, " +
                    "$less<$key_type>>, $allocator<std::pair<$key_type, " +
                    "$mapped_type>>>")
        elif len(c_args) == 5:
            default_hash = 'hash'
            if self.unordered_maps_and_sets:
                tmpl = string.Template(
                    "$container<$key_type, $mapped_type, " +
                    "$hash<$key_type>, " +
                    "$equal_to<$key_type>, " +
                    "$allocator<std::pair<const$key_type, " +
                    "$mapped_type>>>")
                if key_type.startswith('const ') or \
                        key_type.endswith(' const'):
                    tmpl = string.Template(
                        "$container<$key_type, $mapped_type, " +
                        "$hash<$key_type>, " +
                        "$equal_to<$key_type>, " +
                        "$allocator<std::pair<$key_type, " +
                        "$mapped_type>>>")
            else:
                tmpl = string.Template(
                    "$container<$key_type, $mapped_type, "
                    "$hash<$key_type >, " +
                    "$equal_to<$key_type>, "
                    "$allocator<$mapped_type>>")
                if key_type.startswith('const ') or \
                        key_type.endswith(' const'):
                    # TODO: this template is the same than above.
                    # Make sure why this was needed and if this is
                    # tested. There may be a const missing somewhere.
                    tmpl = string.Template(
                        "$container<$key_type, $mapped_type, " +
                        "$hash<$key_type>, " +
                        "$equal_to<$key_type>, " +
                        "$allocator<$mapped_type>>")
        else:
            return

        for ns in std_namespaces:
            inst = tmpl.substitute(
                container=c_name,
                key_type=key_type,
                mapped_type=mapped_type,
                hash=ns + '::' + utils.get_tr1(cls_name) + default_hash,
                less=default_less_or_hash,
                equal_to=default_equal_to,
                allocator=default_allocator)

            if self.normalize(cls_name) == self.normalize(inst):
                return templates.join(
                    c_name,
                    [self.erase_recursive(key_type),
                     self.erase_recursive(mapped_type)])


class container_traits_impl_t(object):

    """
    Implements the functionality needed for convenient work with STD container
    classes

    Implemented functionality:
      * find out whether a declaration is STD container or not
      * find out container value( mapped ) type

    This class tries to be useful as much as possible. For example, for class
    declaration (and not definition) it parses the class name in order to
    extract the information.

    """

    def __init__(
            self,
            container_name,
            element_type_index,
            element_type_typedef,
            eraser,
            key_type_index=None,
            key_type_typedef=None,
            unordered_maps_and_sets=False):
        """
        :param container_name: std container name
        :param element_type_index: position of value\\mapped type within
        template arguments list
        :param element_type_typedef: class typedef to the value\\mapped type
        :param key_type_index: position of key type within template arguments
        list
        :param key_type_typedef: class typedef to the key type
        """
        self._name = container_name
        self.element_type_index = element_type_index
        self.element_type_typedef = element_type_typedef
        self.key_type_index = key_type_index
        self.key_type_typedef = key_type_typedef
        self.unordered_maps_and_sets = unordered_maps_and_sets

        # Get the method from defaults_eraser using it's name
        self.remove_defaults_impl = getattr(
            defaults_eraser(unordered_maps_and_sets), eraser)

    def name(self):
        return self._name

    def get_container_or_none(self, type_):
        """
        Returns reference to the class declaration or None.

        """

        type_ = type_traits.remove_alias(type_)
        type_ = type_traits.remove_cv(type_)

        utils.loggers.queries_engine.debug(
            "Container traits: cleaned up search %s", type_)

        if isinstance(type_, cpptypes.declarated_t):
            cls_declaration = type_traits.remove_alias(type_.declaration)
        elif isinstance(type_, class_declaration.class_t):
            cls_declaration = type_
        elif isinstance(type_, class_declaration.class_declaration_t):
            cls_declaration = type_
        else:
            utils.loggers.queries_engine.debug(
                "Container traits: returning None, type not known\n")
            return

        if not cls_declaration.name.startswith(self.name() + '<'):
            utils.loggers.queries_engine.debug(
                "Container traits: returning None, " +
                "declaration starts with " + self.name() + '<\n')
            return

        # When using libstd++, some container traits are defined in
        # std::tr1::. See remove_template_defaults_tester.py.
        # In this case the is_defined_in_xxx test needs to be done
        # on the parent
        decl = cls_declaration
        if isinstance(type_, class_declaration.class_declaration_t):
            is_ns = isinstance(type_.parent, namespace.namespace_t)
            if is_ns and type_.parent.name == "tr1":
                decl = cls_declaration.parent
        elif isinstance(type_, cpptypes.declarated_t):
            is_ns = isinstance(type_.declaration.parent, namespace.namespace_t)
            if is_ns and type_.declaration.parent.name == "tr1":
                decl = cls_declaration.parent

        for ns in std_namespaces:
            if traits_impl_details.impl_details.is_defined_in_xxx(ns, decl):
                utils.loggers.queries_engine.debug(
                    "Container traits: get_container_or_none() will return " +
                    cls_declaration.name)
                # The is_defined_in_xxx check is done on decl, but we return
                # the original declation so that the rest of the algorithm
                # is able to work with it.
                return cls_declaration

        # This should not happen
        utils.loggers.queries_engine.debug(
            "Container traits: get_container_or_none() will return None\n")

    def is_my_case(self, type_):
        """
        Checks, whether type is STD container or not.

        """

        return bool(self.get_container_or_none(type_))

    def class_declaration(self, type_):
        """
        Returns reference to the class declaration.

        """

        utils.loggers.queries_engine.debug(
            "Container traits: searching class declaration for %s", type_)

        cls_declaration = self.get_container_or_none(type_)
        if not cls_declaration:
            raise TypeError(
                'Type "%s" is not instantiation of std::%s' %
                (type_.decl_string, self.name()))
        return cls_declaration

    def is_sequence(self, type_):
        # raise exception if type is not container
        self.class_declaration(type_)
        return self.key_type_index is None

    def is_mapping(self, type_):
        return not self.is_sequence(type_)

    def __find_xxx_type(
            self,
            type_,
            xxx_index,
            xxx_typedef,
            cache_property_name):
        cls_declaration = self.class_declaration(type_)
        result = getattr(cls_declaration.cache, cache_property_name)
        if not result:
            if isinstance(cls_declaration, class_declaration.class_t):
                xxx_type = cls_declaration.typedef(
                    xxx_typedef, recursive=False).decl_type
                result = type_traits.remove_declarated(xxx_type)
            else:
                xxx_type_str = templates.args(cls_declaration.name)[xxx_index]
                result = traits_impl_details.impl_details.find_value_type(
                    cls_declaration.top_parent, xxx_type_str)
                if None is result:
                    raise RuntimeError(
                        "Unable to find out %s '%s' key\\value type." %
                        (self.name(), cls_declaration.decl_string))
            setattr(cls_declaration.cache, cache_property_name, result)
        return result

    def element_type(self, type_):
        """returns reference to the class value\\mapped type declaration"""
        return self.__find_xxx_type(
            type_,
            self.element_type_index,
            self.element_type_typedef,
            'container_element_type')

    def key_type(self, type_):
        """returns reference to the class key type declaration"""
        if not self.is_mapping(type_):
            raise TypeError(
                'Type "%s" is not "mapping" container' %
                str(type_))
        return self.__find_xxx_type(
            type_,
            self.key_type_index,
            self.key_type_typedef,
            'container_key_type')

    def remove_defaults(self, type_or_string):
        """
        Removes template defaults from a templated class instantiation.

        For example:
            .. code-block:: c++

               std::vector< int, std::allocator< int > >

        will become:
            .. code-block:: c++

               std::vector< int >

        """

        name = type_or_string
        if not isinstance(type_or_string, str):
            name = self.class_declaration(type_or_string).name
        name = defaults_eraser.replace_basic_string(name)
        if not self.remove_defaults_impl:
            return name
        no_defaults = self.remove_defaults_impl(name)
        if not no_defaults:
            return name
        return no_defaults


list_traits = container_traits_impl_t(
    'list',
    0,
    'value_type',
    'erase_allocator')

deque_traits = container_traits_impl_t(
    'deque',
    0,
    'value_type',
    'erase_allocator')

queue_traits = container_traits_impl_t(
    'queue',
    0,
    'value_type',
    'erase_container')

priority_queue_traits = container_traits_impl_t(
    'priority_queue',
    0,
    'value_type',
    'erase_container_compare')

vector_traits = container_traits_impl_t(
    'vector',
    0,
    'value_type',
    'erase_allocator')

stack_traits = container_traits_impl_t(
    'stack',
    0,
    'value_type',
    'erase_container')

map_traits = container_traits_impl_t(
    'map',
    1,
    'mapped_type',
    'erase_map_compare_allocator',
    key_type_index=0,
    key_type_typedef='key_type')

multimap_traits = container_traits_impl_t(
    'multimap',
    1,
    'mapped_type',
    'erase_map_compare_allocator',
    key_type_index=0,
    key_type_typedef='key_type')


hash_map_traits = container_traits_impl_t(
    'hash_map',
    1,
    'mapped_type',
    'erase_hashmap_compare_allocator',
    key_type_index=0,
    key_type_typedef='key_type')


hash_multimap_traits = container_traits_impl_t(
    'hash_multimap',
    1,
    'mapped_type',
    'erase_hashmap_compare_allocator',
    key_type_index=0,
    key_type_typedef='key_type')

set_traits = container_traits_impl_t(
    'set',
    0,
    'value_type',
    'erase_compare_allocator')

multiset_traits = container_traits_impl_t(
    'multiset',
    0,
    'value_type',
    'erase_compare_allocator')

hash_set_traits = container_traits_impl_t(
    'hash_set',
    0,
    'value_type',
    'erase_hash_allocator')

hash_multiset_traits = container_traits_impl_t(
    'hash_multiset',
    0,
    'value_type',
    'erase_hash_allocator')

# c++11 equivalents for clang
unordered_map_traits = container_traits_impl_t(
    'unordered_map',
    1,
    'mapped_type',
    'erase_hashmap_compare_allocator',
    key_type_index=0,
    key_type_typedef='key_type',
    unordered_maps_and_sets=True)

unordered_multimap_traits = container_traits_impl_t(
    'unordered_multimap',
    1,
    'mapped_type',
    'erase_hashmap_compare_allocator',
    key_type_index=0,
    key_type_typedef='key_type',
    unordered_maps_and_sets=True)

unordered_set_traits = container_traits_impl_t(
    'unordered_set',
    0,
    'value_type',
    'erase_hash_allocator',
    unordered_maps_and_sets=True)

unordered_multiset_traits = container_traits_impl_t(
    'unordered_multiset',
    0,
    'value_type',
    'erase_hash_allocator',
    unordered_maps_and_sets=True)

all_container_traits = (
    list_traits,
    deque_traits,
    queue_traits,
    priority_queue_traits,
    vector_traits,
    stack_traits,
    map_traits,
    multimap_traits,
    hash_map_traits,
    hash_multimap_traits,
    set_traits,
    hash_set_traits,
    multiset_traits,
    hash_multiset_traits,
    unordered_map_traits,
    unordered_multimap_traits,
    unordered_set_traits,
    unordered_multiset_traits)
"""tuple of all STD container traits classes"""

sequential_container_traits = [
    list_traits,
    deque_traits,
    queue_traits,
    priority_queue_traits,
    vector_traits,
    stack_traits,
    set_traits,
    hash_set_traits,
    multiset_traits,
    hash_multiset_traits]
"""list, that contains all STD container traits classes"""


def find_container_traits(cls_or_string):
    """
    Find the container traits type of a declaration.

    Args:
        cls_or_string (str | declarations.declaration_t): a string

    Returns:
        declarations.container_traits: a container traits
    """

    if isinstance(cls_or_string, str):
        if not templates.is_instantiation(cls_or_string):
            return None
        name = templates.name(cls_or_string)
        if name.startswith('std::'):
            name = name[len('std::'):]
        if name.startswith('std::tr1::'):
            name = name[len('std::tr1::'):]
        for cls_traits in all_container_traits:
            if cls_traits.name() == name:
                return cls_traits
    else:

        if isinstance(cls_or_string, class_declaration.class_types):
            # Look in the cache.
            if cls_or_string.cache.container_traits is not None:
                return cls_or_string.cache.container_traits

        # Look for a container traits
        for cls_traits in all_container_traits:
            if cls_traits.is_my_case(cls_or_string):
                # Store in the cache
                if isinstance(cls_or_string, class_declaration.class_types):
                    cls_or_string.cache.container_traits = cls_traits
                return cls_traits



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Contains enumeration of all xml_generators supported by the project.

"""

gccxml_06 = "GCC-XML 0.6"
gccxml_07 = "GCC-XML 0.7"
gccxml_09 = "GCC-XML 0.9"
gccxml_09_buggy = "GCC-XML 0.9 BUGGY"
# revision 122:
# After this fix, all constructors and destructors that exist for a class
# are dumped whether the user declared them or not.  Those that were
# implicitly declared by the xml_generator are marked as "artificial".

# CastXML has no version number for the moment so "None" is used.
castxml_none = "CastXML None"


def on_missing_functionality(xml_generator, functionality):
    raise NotImplementedError(
        '"%s" xml_generator doesn\'t support functionality "%s"' %
        (xml_generator, functionality))



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""Python CastXML front end.

This package provides functionality to extract and inspect
declarations from C/C++ header files. This is accomplished
by invoking an external tool like CastXML,
which parses a header file and dumps the declarations as a
XML file. This XML file is then read by pygccxml and the contents
are made available as appropriate Python objects.

To parse a set of C/C++ header files you use the :func:`parse <parser.parse>`
function in the :mod:parser sub package which returns a tree that contains all
declarations found in the header files. The root of the tree represents the
main namespace `::` and the children nodes represent the namespace contents
such as other namespaces, classes, functions, etc. Each node in the tree is an
object of a type derived from the :class:`declaration_t` class. An inner node
is always either a namespace :class:`declarations.namespace_t` or a class
:class:`declarations.class_t`, which are both derived from
:class:`declarations.scopedef_t` class. Everything else (free functions,
member functions, enumerations, variables, etc.) are always a leaf. You will
find all those declaration classes in the :mod:declarations sub-package.

"""

import sys
import warnings

from . import declarations
from . import parser
from . import utils

# Always show deprecation warnings.
# These are hidden by default since python 2.7.
# pygccxml is a tool for developers, and these need
# to know what is deprecated.
warnings.simplefilter("always", DeprecationWarning)

version = sys.version_info

if version < (3, 8):
    import importlib_metadata as metadata
elif version < (3, 9, 10) or (3, 10, 0) <= version < (3, 10, 2):
    try:
        import importlib_metadata as metadata
    except ModuleNotFoundError:
        from importlib import metadata
else:
    from importlib import metadata

__version__ = metadata.version("pygccxml")



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Defines C++ parser configuration classes.

"""

import os
import copy
import platform
import shutil
import subprocess
import warnings
# In py3, ConfigParser was renamed to the more-standard configparser.
# But there's a py3 backport that installs "configparser" in py2, and I don't
# want it because it has annoying deprecation warnings. So try the real py2
# import first
# Inspired by https://bitbucket.org/ned/coveragepy/commits/f8e9d62f1412
try:
    from ConfigParser import SafeConfigParser as ConfigParser
except ImportError:
    from configparser import ConfigParser


class parser_configuration_t(object):

    """
    C++ parser configuration holder

    This class serves as a base class for the parameters that can be used
    to customize the call to a C++ parser.

    This class also allows users to work with relative files paths. In this
    case files are searched in the following order:

       1. current directory
       2. working directory
       3. additional include paths specified by the user

    """

    def __init__(
            self,
            working_directory='.',
            include_paths=None,
            define_symbols=None,
            undefine_symbols=None,
            cflags="",
            ccflags="",
            compiler=None,
            xml_generator=None,
            keep_xml=False,
            compiler_path=None,
            flags=None,
            castxml_epic_version=None):

        object.__init__(self)
        self.__working_directory = working_directory

        if not include_paths:
            include_paths = []
        self.__include_paths = include_paths

        if not define_symbols:
            define_symbols = []
        self.__define_symbols = define_symbols

        if not undefine_symbols:
            undefine_symbols = []
        self.__undefine_symbols = undefine_symbols

        self.__cflags = cflags

        self.__ccflags = ccflags

        self.__compiler = compiler

        self.__xml_generator = xml_generator

        self.__castxml_epic_version = castxml_epic_version

        self.__keep_xml = keep_xml

        if flags is None:
            flags = []
        self.__flags = flags

        # If no compiler path was set and we are using castxml, set the path
        self.__compiler_path = create_compiler_path(
            xml_generator, compiler_path)

    def clone(self):
        raise NotImplementedError(self.__class__.__name__)

    @property
    def working_directory(self):
        return self.__working_directory

    @working_directory.setter
    def working_directory(self, working_dir):
        self.__working_directory = working_dir

    @property
    def include_paths(self):
        """list of include paths to look for header files"""
        return self.__include_paths

    @property
    def define_symbols(self):
        """list of "define" directives """
        return self.__define_symbols

    @property
    def undefine_symbols(self):
        """list of "undefine" directives """
        return self.__undefine_symbols

    @property
    def compiler(self):
        """get compiler name to simulate"""
        return self.__compiler

    @compiler.setter
    def compiler(self, compiler):
        """set compiler name to simulate"""
        self.__compiler = compiler

    @property
    def xml_generator(self):
        """get xml_generator"""
        return self.__xml_generator

    @xml_generator.setter
    def xml_generator(self, xml_generator):
        """set xml_generator"""
        self.__xml_generator = xml_generator

    @property
    def castxml_epic_version(self):
        """
        File format version used by castxml.
        """
        return self.__castxml_epic_version

    @castxml_epic_version.setter
    def castxml_epic_version(self, castxml_epic_version):
        """
        File format version used by castxml.
        """
        self.__castxml_epic_version = castxml_epic_version

    @property
    def keep_xml(self):
        """Are xml files kept after errors."""
        return self.__keep_xml

    @keep_xml.setter
    def keep_xml(self, keep_xml):
        """Set if xml files kept after errors."""
        self.__keep_xml = keep_xml

    @property
    def flags(self):
        """Optional flags for pygccxml."""
        return self.__flags

    @flags.setter
    def flags(self, flags):
        """Optional flags for pygccxml."""
        if flags is None:
            flags = []
        self.__flags = flags

    @property
    def compiler_path(self):
        """Get the path for the compiler."""
        return self.__compiler_path

    @compiler_path.setter
    def compiler_path(self, compiler_path):
        """Set the path for the compiler."""
        self.__compiler_path = compiler_path

    @property
    def cflags(self):
        """additional flags to pass to compiler"""
        return self.__cflags

    @cflags.setter
    def cflags(self, val):
        self.__cflags = val

    def append_cflags(self, val):
        self.__cflags = self.__cflags + ' ' + val

    @property
    def ccflags(self):
        """
        additional cross-compatible flags to pass directly
        to internal simulated compiler.
        Castxml removes any definitions of its
        pre-defined macros (e.g. -fopenmp). To propagate these down to the
        compiler, these flags must also be passed here.
        See `cc-opt` on castxml's documentation page:
        https://github.com/CastXML/CastXML/blob/master/doc/manual/castxml.1.rst
        """
        return self.__ccflags

    @ccflags.setter
    def ccflags(self, val):
        self.__ccflags = val

    def append_ccflags(self, val):
        self.__ccflags = self.__ccflags + ' ' + val

    def __ensure_dir_exists(self, dir_path, meaning):
        if os.path.isdir(dir_path):
            return
        if os.path.exists(self.working_directory):
            msg = '%s("%s") does not exist.' % (meaning, dir_path)
            if meaning == 'include directory':
                # Warn instead of failing.
                warnings.warn(msg, RuntimeWarning)
            else:
                raise RuntimeError(msg)
        else:
            raise RuntimeError(
                '%s("%s") should be "directory", not a file.' %
                (meaning, dir_path))

    def raise_on_wrong_settings(self):
        """
        Validates the configuration settings and raises RuntimeError on error
        """
        self.__ensure_dir_exists(self.working_directory, 'working directory')
        for idir in self.include_paths:
            self.__ensure_dir_exists(idir, 'include directory')
        if self.__xml_generator != "castxml":
            msg = f"xml_generator ({self.xml_generator}) can only be 'castxml'"
            raise RuntimeError(msg)


class xml_generator_configuration_t(parser_configuration_t):
    """
    Configuration object to collect parameters for invoking gccxml or castxml.

    This class serves as a container for the parameters that can be used
    to customize the call to gccxml or castxml.

    """

    def __init__(
            self,
            gccxml_path='',
            xml_generator_path='',
            working_directory='.',
            include_paths=None,
            define_symbols=None,
            undefine_symbols=None,
            start_with_declarations=None,
            ignore_gccxml_output=False,
            cflags="",
            ccflags="",
            compiler=None,
            xml_generator=None,
            keep_xml=False,
            compiler_path=None,
            flags=None,
            castxml_epic_version=None):

        parser_configuration_t.__init__(
            self,
            working_directory=working_directory,
            include_paths=include_paths,
            define_symbols=define_symbols,
            undefine_symbols=undefine_symbols,
            cflags=cflags,
            ccflags=ccflags,
            compiler=compiler,
            xml_generator=xml_generator,
            keep_xml=keep_xml,
            compiler_path=compiler_path,
            flags=flags,
            castxml_epic_version=castxml_epic_version)

        if gccxml_path != '':
            self.__gccxml_path = gccxml_path
        self.__xml_generator_path = xml_generator_path

        if not start_with_declarations:
            start_with_declarations = []
        self.__start_with_declarations = start_with_declarations

        self.__ignore_gccxml_output = ignore_gccxml_output

        self.__xml_generator_from_xml_file = None

    def clone(self):
        return copy.deepcopy(self)

    @property
    def xml_generator_path(self):
        """
        XML generator binary location

        """

        return self.__xml_generator_path

    @xml_generator_path.setter
    def xml_generator_path(self, new_path):
        self.__xml_generator_path = new_path

    @property
    def xml_generator_from_xml_file(self):
        """
        Configuration object containing information about the xml generator
        read from the xml file.

        Returns:
            utils.xml_generators: configuration object
        """
        return self.__xml_generator_from_xml_file

    @xml_generator_from_xml_file.setter
    def xml_generator_from_xml_file(self, xml_generator_from_xml_file):
        self.__xml_generator_from_xml_file = xml_generator_from_xml_file

    @property
    def start_with_declarations(self):
        """list of declarations gccxml should start with, when it dumps
        declaration tree"""
        return self.__start_with_declarations

    @property
    def ignore_gccxml_output(self):
        """set this property to True, if you want pygccxml to ignore any
            error warning that comes from gccxml"""
        return self.__ignore_gccxml_output

    @ignore_gccxml_output.setter
    def ignore_gccxml_output(self, val=True):
        self.__ignore_gccxml_output = val

    def raise_on_wrong_settings(self):
        super(xml_generator_configuration_t, self).raise_on_wrong_settings()
        if self.xml_generator_path is None or \
                not os.path.isfile(self.xml_generator_path):
            msg = (
                'xml_generator_path("%s") should be set and exist.') \
                % self.xml_generator_path
            raise RuntimeError(msg)


def load_xml_generator_configuration(configuration, **defaults):
    """
    Loads CastXML or GCC-XML configuration.

    Args:
         configuration (string|configparser.ConfigParser): can be
             a string (file path to a configuration file) or
             instance of :class:`configparser.ConfigParser`.
         defaults: can be used to override single configuration values.

    Returns:
        :class:`.xml_generator_configuration_t`: a configuration object


    The file passed needs to be in a format that can be parsed by
    :class:`configparser.ConfigParser`.

    An example configuration file skeleton can be found
    `here <https://github.com/gccxml/pygccxml/blob/develop/
    tests/xml_generator.cfg>`_.

    """
    parser = configuration
    if isinstance(configuration, str):
        parser = ConfigParser()
        parser.read(configuration)

    # Create a new empty configuration
    cfg = xml_generator_configuration_t()

    values = defaults
    if not values:
        values = {}

    if parser.has_section('xml_generator'):
        for name, value in parser.items('xml_generator'):
            if value.strip():
                values[name] = value

    for name, value in values.items():
        if isinstance(value, str):
            value = value.strip()
        if name == 'gccxml_path':
            cfg.gccxml_path = value
        if name == 'xml_generator_path':
            cfg.xml_generator_path = value
        elif name == 'working_directory':
            cfg.working_directory = value
        elif name == 'include_paths':
            for p in value.split(';'):
                p = p.strip()
                if p:
                    cfg.include_paths.append(os.path.normpath(p))
        elif name == 'compiler':
            cfg.compiler = value
        elif name == 'xml_generator':
            cfg.xml_generator = value
        elif name == 'castxml_epic_version':
            cfg.castxml_epic_version = int(value)
        elif name == 'keep_xml':
            cfg.keep_xml = value
        elif name == 'cflags':
            cfg.cflags = value
        elif name == 'ccflags':
            cfg.ccflags = value
        elif name == 'flags':
            cfg.flags = value
        elif name == 'compiler_path':
            cfg.compiler_path = value
        else:
            print('\n%s entry was ignored' % name)

    # If no compiler path was set and we are using castxml, set the path
    # Here we overwrite the default configuration done in the cfg because
    # the xml_generator was set through the setter after the creation of a new
    # emppty configuration object.
    cfg.compiler_path = create_compiler_path(
        cfg.xml_generator, cfg.compiler_path)

    return cfg


def create_compiler_path(xml_generator, compiler_path):
    """
    Try to guess a path for the compiler.

    If you want to use a specific compiler, please provide the compiler
    path manually, as the guess may not be what you are expecting.
    Providing the path can be done by passing it as an argument (compiler_path)
    to the xml_generator_configuration_t() or by defining it in your pygccxml
    configuration file.

    """

    if xml_generator == 'castxml' and compiler_path is None:
        if platform.system() == 'Windows':
            # Look for msvc
            compiler_path = shutil.which('cl')
            # No msvc found; look for mingw
            if compiler_path is None:
                compiler_path = shutil.which('mingw')
        else:
            # OS X or Linux
            # Look for clang first, then gcc
            compiler_path = shutil.which('clang++')
            # No clang found; use gcc
            if compiler_path is None:
                compiler_path = shutil.which('c++')

    return compiler_path


if __name__ == '__main__':
    print(load_xml_generator_configuration('xml_generator.cfg').__dict__)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Defines classes that will keep results of different calculations.

"""


class declaration_algs_cache_t(object):

    def __init__(self):
        object.__init__(self)
        self._enabled = True
        self._full_name = None
        self._full_partial_name = None
        self._access_type = None
        self._declaration_path = None
        self._partial_declaration_path = None
        self._container_key_type = None
        self._container_element_type = None
        self._cmp_data = None
        self._normalized_name = None
        self._normalized_partial_name = None
        self._normalized_full_name_true = None
        self._normalized_full_name_false = None
        self._container_traits = None

    def disable(self):
        self._enabled = False

    def enable(self):
        self._enabled = True

    @property
    def enabled(self):
        return self._enabled

    @property
    def full_name(self):
        return self._full_name

    @full_name.setter
    def full_name(self, fname):
        if not self.enabled:
            fname = None
        self._full_name = fname

    @property
    def container_traits(self):
        return self._container_traits

    @container_traits.setter
    def container_traits(self, container_traits):
        if not self.enabled:
            container_traits = None
        self._container_traits = container_traits

    @property
    def full_partial_name(self):
        return self._full_partial_name

    @full_partial_name.setter
    def full_partial_name(self, fname):
        if not self.enabled:
            fname = None
        self._full_partial_name = fname

    @property
    def access_type(self):
        return self._access_type

    @access_type.setter
    def access_type(self, access_type):
        if not self.enabled:
            access_type = None
        self._access_type = access_type

    @property
    def declaration_path(self):
        return self._declaration_path

    @declaration_path.setter
    def declaration_path(self, declaration_path):
        if not self.enabled:
            declaration_path = None
        self._declaration_path = declaration_path

    @property
    def partial_declaration_path(self):
        return self._partial_declaration_path

    @partial_declaration_path.setter
    def partial_declaration_path(self, partial_declaration_path):
        if not self.enabled:
            partial_declaration_path = None
        self._partial_declaration_path = partial_declaration_path

    @property
    def container_element_type(self):
        return self._container_element_type

    @container_element_type.setter
    def container_element_type(self, etype):
        if not self.enabled:
            etype = None
        self._container_element_type = etype

    @property
    def container_key_type(self):
        return self._container_key_type

    @container_key_type.setter
    def container_key_type(self, ktype):
        if not self.enabled:
            ktype = None
        self._container_key_type = ktype

    @property
    def normalized_name(self):
        return self._normalized_name

    @normalized_name.setter
    def normalized_name(self, normalized_name):
        if not self.enabled:
            normalized_name = None
        self._normalized_name = normalized_name

    @property
    def normalized_partial_name(self):
        return self._normalized_partial_name

    @normalized_partial_name.setter
    def normalized_partial_name(self, normalized_partial_name):
        if not self.enabled:
            normalized_partial_name = None
        self._normalized_partial_name = normalized_partial_name

    @property
    def normalized_full_name_true(self):
        return self._normalized_full_name_true

    @normalized_full_name_true.setter
    def normalized_full_name_true(self, normalized_full_name_true):
        if not self.enabled:
            normalized_full_name_true = None
        self._normalized_full_name_true = normalized_full_name_true

    @property
    def normalized_full_name_false(self):
        return self._normalized_full_name_false

    @normalized_full_name_false.setter
    def normalized_full_name_false(self, normalized_full_name_false):
        if not self.enabled:
            normalized_full_name_false = None
        self._normalized_full_name_false = normalized_full_name_false

    @property
    def cmp_data(self):
        """Data used for comparison between declarations."""
        return self._cmp_data

    @cmp_data.setter
    def cmp_data(self, cmp_data):
        """Data used for comparison between declarations."""
        if not self.enabled:
            cmp_data = None
        self._cmp_data = cmp_data

    def reset(self):
        self.full_name = None
        self.full_partial_name = None
        self.access_type = None
        self.declaration_path = None
        self.partial_declaration_path = None
        self.container_key_type = None
        self.container_element_type = None
        self.cmp_data = None
        self.normalized_name = None
        self.normalized_partial_name = None
        self.normalized_full_name_true = None
        self.normalized_full_name_false = None
        self.container_traits = None

    def reset_name_based(self):
        self.full_name = None
        self.full_partial_name = None
        self.declaration_path = None
        self.partial_declaration_path = None
        self.container_key_type = None
        self.container_element_type = None
        self.cmp_data = None
        self.normalized_name = None
        self.normalized_partial_name = None
        self.normalized_full_name_true = None
        self.normalized_full_name_false = None
        self.container_traits = None

    def reset_access_type(self):
        self.access_type = None


class type_algs_cache_t(object):
    enabled = True

    @staticmethod
    def disable():
        type_algs_cache_t.enabled = False

    @staticmethod
    def enable():
        type_algs_cache_t.enabled = True

    def __init__(self):
        object.__init__(self)
        self._remove_alias = None
        self._decl_string = None
        self._partial_decl_string = None

    @property
    def remove_alias(self):
        return self._remove_alias

    @remove_alias.setter
    def remove_alias(self, remove_alias):
        if not type_algs_cache_t.enabled:
            remove_alias = None
        self._remove_alias = remove_alias

    @property
    def decl_string(self):
        return self._decl_string

    @decl_string.setter
    def decl_string(self, decl_string):
        if not type_algs_cache_t.enabled:
            decl_string = None
        self._decl_string = decl_string

    @property
    def partial_decl_string(self):
        return self._partial_decl_string

    @partial_decl_string.setter
    def partial_decl_string(self, partial_decl_string):
        if not type_algs_cache_t.enabled:
            partial_decl_string = None
        self._partial_decl_string = partial_decl_string

    def reset(self):
        self.remove_alias = None
        self.decl_string = None
        self.partial_decl_string = None



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Describe a C++ comment declaration.

"""

from . import location as pygccxml_location


class comment_t(object):

    def __init__(self, name='', declarations=None):
        """
        Creates an object that describes a C++ comment declaration.

        Args:

        """
        self._location = {}
        self._begin_line = 0
        self._begin_column = 0
        self._begin_offset = 0
        self._end_line = 0
        self._end_column = 0
        self._end_offset = 0
        self._text = ""

    @property
    def location(self):
        """An instance of the location_t class
        which contains the file where the
        comment can be found.
        @type: location_t """
        return self._location

    @location.setter
    def location(self, location):
        if not isinstance(location, pygccxml_location.location_t):
            raise ValueError(
                "'location' must be a location_t (got a %s instead)" %
                type(location).__name__)
        self._location = location

    @property
    def begin_line(self):
        """An integer value which corresponds to the
        line of the file where the comment begins
        @type: int """
        return self._begin_line

    @begin_line.setter
    def begin_line(self, begin_line):
        self._begin_line = int(begin_line)

    @property
    def begin_offset(self):
        """An integer value representing the
        number of bytes from the beginning of the
        file to the start of the comment
        @type: int """
        return self._begin_offset

    @begin_offset.setter
    def begin_offset(self, begin_offset):
        self._begin_offset = int(begin_offset)

    @property
    def begin_column(self):
        """An integer value which corresponds to the
        column of the file where the comment begins
        @type: int """
        return self._begin_column

    @begin_column.setter
    def begin_column(self, begin_column):
        self._begin_column = int(begin_column)

    @property
    def end_line(self):
        """An integer value which corresponds to the
        line of the file where the comment ends
        @type: int """
        return self._end_line

    @end_line.setter
    def end_line(self, end_line):
        self._end_line = int(end_line)

    @property
    def end_offset(self):
        """An integer value representing the
        number of bytes from the beginning of the
        file to the end of the comment
        @type: int """
        return self._end_offset

    @end_offset.setter
    def end_offset(self, end_offset):
        self._end_offset = int(end_offset)

    @property
    def end_column(self):
        """An integer value which corresponds to the
        column of character in a line of the file
        where the comment ends
        @type: int """
        return self._end_column

    @end_column.setter
    def end_column(self, end_column):
        self._end_column = int(end_column)

    @property
    def text(self):
        """A list of strings where each entry in the list
        is one line of the comment. These comments will not
        end in a newline
        @type: list """
        return self._text

    @text.setter
    def text(self, text):
        if not isinstance(text, list):
            raise ValueError(
                "'text' must be a list (got a %s instead)" %
                type(text).__name__)
        self._text = text



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt


class location_t(object):
    """
    Provides information about the location of the declaration within the
    source file.

    """

    def __init__(self, file_name='', line=-1):
        self._file_name = file_name
        self._line = line

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.line == other.line \
            and self.file_name == other.file_name

    def __hash__(self):
        return hash(self.__class__) ^ hash(self.line) ^ hash(self.file_name)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, location_t):
            return self.__class__.__name__ < other.__class__.__name__
        return (self.file_name, self.line) < (other.file_name, other.line)

    @property
    def file_name(self):
        """
        Absolute source file name, type string.

        """

        return self._file_name

    @file_name.setter
    def file_name(self, new_file_name):
        self._file_name = new_file_name

    @property
    def line(self):
        """
        Line number, type int.

        """

        return self._line

    @line.setter
    def line(self, new_line):
        self._line = new_line

    def as_tuple(self):
        """
        Return tuple(self.file_name, self.line)

        """

        return self.file_name, self.line



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

import os

from . import templates
from . import matcher_base_t
from . import variable
from . import cpptypes
from . import namespace
from . import calldef
from . import calldef_members
from .. import utils


class declaration_matcher_t(matcher_base_t):

    """
    Instance of this class will match declarations by next criteria:
          - declaration name, also could be fully qualified name
            Example: `wstring` or `::std::wstring`
          - declaration type
            Example: :class:`class_t`, :class:`namespace_t`,
            :class:`enumeration_t`
          - location within file system ( file or directory )
    """

    def __init__(
            self,
            name=None,
            decl_type=None,
            header_dir=None,
            header_file=None):
        """
        :param decl_type: declaration type to match by. For example
        :class:`enumeration_t`.
        :type decl_type: any class that derives from :class:`declaration_t`
        class

        :param name: declaration name, could be full name.
        :type name: str

        :param header_dir: absolute directory path
        :type header_dir: str

        :param header_file: absolute file path
        :type header_file: str

        """
        # An other option is that pygccxml will create absolute path using
        # os.path.abspath function. But I think this is just wrong, because
        # abspath builds path using current working directory. This behavior
        # is fragile and very difficult to find a bug.
        matcher_base_t.__init__(self)
        self.decl_type = decl_type
        self.__name = None
        self.__opt_is_tmpl_inst = None
        self.__opt_tmpl_name = None
        self.__opt_is_full_name = None
        self.__decl_name_only = None

        # Set the name through the setter.
        self.name = name

        self.header_dir = header_dir
        self.header_file = header_file

        if self.header_dir:
            self.header_dir = utils.normalize_path(self.header_dir)
            if not os.path.isabs(self.header_dir):
                raise RuntimeError(
                    "Path to header directory should be absolute!")

        if self.header_file:
            self.header_file = utils.normalize_path(self.header_file)
            if not os.path.isabs(self.header_file):
                raise RuntimeError("Path to header file should be absolute!")

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        self.__name = name
        if not self.__name:
            self.__opt_is_tmpl_inst = None
            self.__opt_tmpl_name = None
            self.__opt_is_full_name = None
            self.__decl_name_only = None
        else:
            self.__opt_is_tmpl_inst = templates.is_instantiation(self.__name)
            self.__opt_tmpl_name = templates.name(self.__name)
            if self.__opt_is_tmpl_inst:
                if '::' in self.__opt_tmpl_name:
                    self.__opt_is_full_name = True
                    self.__decl_name_only = \
                        self.__opt_tmpl_name.split('::')[-1]
                else:
                    self.__opt_is_full_name = False
                    self.__decl_name_only = self.__opt_tmpl_name
                self.__name = templates.normalize(name)
            else:
                if '::' in self.__name:
                    self.__opt_is_full_name = True
                    self.__decl_name_only = self.__name.split('::')[-1]
                else:
                    self.__opt_is_full_name = False
                    self.__decl_name_only = self.__name

    def __str__(self):
        msg = []
        if self.decl_type is not None:
            msg.append('(decl type==%s)' % self.decl_type.__name__)
        if self.name is not None:
            msg.append('(name==%s)' % self.name)
        if self.header_dir is not None:
            msg.append('(header dir==%s)' % self.header_dir)
        if self.header_file is not None:
            msg.append('(header file==%s)' % self.header_file)
        if not msg:
            msg.append('any')
        return ' and '.join(msg)

    def __call__(self, decl):
        if self.decl_type is not None:
            if not isinstance(decl, self.decl_type):
                return False
        if self.name is not None:
            if not self.check_name(decl):
                return False
        if self.header_dir is not None:
            if decl.location:
                decl_dir = os.path.abspath(
                    os.path.dirname(decl.location.file_name))
                decl_dir = utils.normalize_path(decl_dir)
                if decl_dir[:len(self.header_dir)] != self.header_dir:
                    return False
            else:
                return False
        if self.header_file is not None:
            if decl.location:
                decl_file = os.path.abspath(decl.location.file_name)
                decl_file = utils.normalize_path(decl_file)
                if decl_file != self.header_file:
                    return False
            else:
                return False
        return True

    def check_name(self, decl):
        assert self.name is not None
        if self.__opt_is_tmpl_inst:
            if not self.__opt_is_full_name:
                if self.name != templates.normalize_name(decl) \
                   and self.name != templates.normalize_partial_name(decl):
                    return False
            else:
                if self.name != templates.normalize_full_name_true(decl) and \
                        self.name != templates.normalize_full_name_false(decl):
                    return False
        else:
            if not self.__opt_is_full_name:
                if self.name != decl.name and self.name != decl.partial_name:
                    return False
            else:
                if self.name != templates.normalize_full_name_true(decl) and \
                        self.name != templates.normalize_full_name_false(decl):
                    return False
        return True

    def is_full_name(self):
        return self.__opt_is_full_name

    @property
    def decl_name_only(self):
        return self.__decl_name_only


class variable_matcher_t(declaration_matcher_t):

    """
    Instance of this class will match variables by next criteria:
        - :class:`declaration_matcher_t` criteria
        - variable type. Example: :class:`int_t` or 'int'
    """

    def __init__(
            self,
            name=None,
            decl_type=None,
            header_dir=None,
            header_file=None):
        """
        :param decl_type: variable type
        :type decl_type: string or instance of :class:`type_t` derived class
        """

        declaration_matcher_t.__init__(
            self,
            name=name,
            decl_type=variable.variable_t,
            header_dir=header_dir,
            header_file=header_file)
        self._decl_type = decl_type

    def __call__(self, decl):
        if not super(variable_matcher_t, self).__call__(decl):
            return False
        if self._decl_type is not None:
            if isinstance(self._decl_type, cpptypes.type_t):
                if self._decl_type != decl.decl_type:
                    return False
            else:
                if self._decl_type != decl.decl_type.decl_string:
                    return False
        return True

    def __str__(self):
        msg = [super(variable_matcher_t, self).__str__()]
        if msg == ['any']:
            msg = []
        if self._decl_type is not None:
            msg.append('(value type==%s)' % str(self._decl_type))
        if not msg:
            msg.append('any')
        return ' and '.join(msg)


class namespace_matcher_t(declaration_matcher_t):

    """Instance of this class will match namespaces by name."""

    def __init__(self, name=None):
        declaration_matcher_t.__init__(
            self,
            name=name,
            decl_type=namespace.namespace_t)

    def __call__(self, decl):
        if self.name and decl.name == '':
            # unnamed namespace have same name as thier parent, we should
            # prevent this happens. The price is: user should search for
            # unnamed namespace directly.
            return False
        return super(namespace_matcher_t, self).__call__(decl)


class calldef_matcher_t(declaration_matcher_t):

    """
    Instance of this class will match callable by the following criteria:
       * :class:`declaration_matcher_t` criteria
       * return type. For example: :class:`int_t` or 'int'
       * argument types

    """

    def __init__(
            self,
            name=None,
            return_type=None,
            arg_types=None,
            decl_type=None,
            header_dir=None,
            header_file=None):
        """
        :param return_type: callable return type
        :type return_type: string or instance of :class:`type_t` derived class

        :type arg_types: list
        :param arg_types: list of function argument types. `arg_types` can
                          contain.
                          Any item within the list could be string or instance
                          of :class:`type_t` derived class. If you don't want
                          some argument to participate in match you can put
                          None.

        For example:

          .. code-block:: python

             calldef_matcher_t( arg_types=[ 'int &', None ] )

        will match all functions that takes 2 arguments, where the first one is
        reference to integer and second any
        """
        if None is decl_type:
            decl_type = calldef.calldef_t
        declaration_matcher_t.__init__(
            self,
            name=name,
            decl_type=decl_type,
            header_dir=header_dir,
            header_file=header_file)

        self.return_type = return_type
        self.arg_types = arg_types

    def __call__(self, decl):
        if not super(calldef_matcher_t, self).__call__(decl):
            return False
        if self.return_type is not None \
           and not self.__compare_types(self.return_type, decl.return_type):
            return False
        if self.arg_types:
            if isinstance(self.arg_types, (list, tuple)):
                if len(self.arg_types) != len(decl.arguments):
                    return False
                for type_or_str, arg in zip(self.arg_types, decl.arguments):
                    if type_or_str is None:
                        continue
                    else:
                        if not self.__compare_types(
                                type_or_str, arg.decl_type):
                            return False
        return True

    @staticmethod
    def __compare_types(type_or_str, decl_type):
        assert type_or_str
        if decl_type is None:
            return False
        if isinstance(type_or_str, cpptypes.type_t):
            if type_or_str != decl_type:
                return False
        else:
            if type_or_str != decl_type.decl_string:
                return False
        return True

    def __str__(self):
        msg = [super(calldef_matcher_t, self).__str__()]
        if msg == ['any']:
            msg = []
        if self.return_type is not None:
            msg.append('(return type==%s)' % str(self.return_type))
        if self.arg_types:
            for i, arg_type in enumerate(self.arg_types):
                if arg_type is None:
                    msg.append('(arg %d type==any)' % i)
                else:
                    msg.append('(arg %d type==%s)' % (i, str(arg_type)))
        if not msg:
            msg.append('any')
        return ' and '.join(msg)


class operator_matcher_t(calldef_matcher_t):

    """
    Instance of this class will match operators by next criteria:
        * :class:`calldef_matcher_t` criteria
        * operator symbol: =, !=, (), [] and etc
    """

    def __init__(
            self,
            name=None,
            symbol=None,
            return_type=None,
            arg_types=None,
            decl_type=None,
            header_dir=None,
            header_file=None):
        """
        :param symbol: operator symbol
        :type symbol: str
        """
        if None is decl_type:
            decl_type = calldef_members.operator_t
        calldef_matcher_t.__init__(
            self,
            name=name,
            return_type=return_type,
            arg_types=arg_types,
            decl_type=decl_type,
            header_dir=header_dir,
            header_file=header_file)
        self.symbol = symbol

    def __call__(self, decl):
        if not super(operator_matcher_t, self).__call__(decl):
            return False
        if self.symbol is not None:
            if self.symbol != decl.symbol:
                return False
        return True

    def __str__(self):
        msg = [super(operator_matcher_t, self).__str__()]
        if msg == ['any']:
            msg = []
        if self.symbol is not None:
            msg.append('(symbol==%s)' % str(self.symbol))
        if not msg:
            msg.append('any')
        return ' and '.join(msg)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
The utils package contains tools used internally by pygccxml.

"""

from .utils import get_architecture
from .utils import loggers
from .utils import create_temp_file_name
from .utils import remove_file_no_raise
from .utils import normalize_path
from .utils import find_xml_generator
from .utils import get_tr1
from .utils import cxx_standard
from .xml_generators import xml_generators



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
Contains classes that describe different C++ declarations

"""

# Keep compilers for backward compatibility
from . import xml_generators as compilers
from . import xml_generators

from .location import location_t
from .declaration import declaration_t
from .scopedef import scopedef_t
from .enumeration import enumeration_t
from .typedef import typedef_t

from .namespace import namespace_t
from .namespace import get_global_namespace

from .class_declaration import class_t
from .class_declaration import CLASS_TYPES
from .class_declaration import ACCESS_TYPES
from .class_declaration import hierarchy_info_t
from .class_declaration import class_declaration_t
from .class_declaration import class_types

from .cpptypes import type_t
from .cpptypes import dummy_type_t
from .cpptypes import unknown_t
from .cpptypes import ellipsis_t
from .cpptypes import fundamental_t
from .cpptypes import void_t
from .cpptypes import char_t
from .cpptypes import signed_char_t
from .cpptypes import unsigned_char_t
from .cpptypes import wchar_t
from .cpptypes import short_int_t
from .cpptypes import short_unsigned_int_t
from .cpptypes import bool_t
from .cpptypes import int_t
from .cpptypes import unsigned_int_t
from .cpptypes import long_int_t
from .cpptypes import long_unsigned_int_t
from .cpptypes import long_long_int_t
from .cpptypes import long_long_unsigned_int_t
from .cpptypes import int128_t
from .cpptypes import uint128_t
from .cpptypes import float_t
from .cpptypes import double_t
from .cpptypes import long_double_t
from .cpptypes import FUNDAMENTAL_TYPES
from .cpptypes import compound_t
from .cpptypes import volatile_t
from .cpptypes import const_t
from .cpptypes import pointer_t
from .cpptypes import reference_t
from .cpptypes import elaborated_t
from .cpptypes import restrict_t
from .cpptypes import array_t
from .cpptypes import calldef_type_t
from .cpptypes import free_function_type_t
from .cpptypes import member_function_type_t
from .cpptypes import member_variable_type_t
from .cpptypes import declarated_t
from .cpptypes import type_qualifiers_t
# java types
from .cpptypes import java_fundamental_t
from .cpptypes import jbyte_t
from .cpptypes import jshort_t
from .cpptypes import jint_t
from .cpptypes import jlong_t
from .cpptypes import jfloat_t
from .cpptypes import jdouble_t
from .cpptypes import jchar_t
from .cpptypes import jboolean_t

from .variable import variable_t

from .declaration_utils import full_name
from .declaration_utils import full_name_from_declaration_path
from .declaration_utils import declaration_path
from .declaration_utils import get_named_parent

from .scopedef import make_flatten
from .scopedef import find_declaration
from .scopedef import find_all_declarations
from .scopedef import find_first_declaration
from .scopedef import declaration_files
from .scopedef import matcher

from .algorithm import apply_visitor
from .algorithm import match_declaration_t

from .calldef_types import VIRTUALITY_TYPES
from .calldef_types import FUNCTION_VIRTUALITY_TYPES
from .calldef_types import CALLING_CONVENTION_TYPES

from .calldef import argument_t
from .calldef import calldef_t
from .calldef_members import member_calldef_t
from .calldef_members import operator_t
from .calldef_members import member_function_t
from .calldef_members import constructor_t
from .calldef_members import destructor_t
from .calldef_members import member_operator_t
from .calldef_members import casting_operator_t

from .free_calldef import free_calldef_t
from .free_calldef import free_function_t
from .free_calldef import free_operator_t

from .decl_visitor import decl_visitor_t
from .type_visitor import type_visitor_t

from .type_traits import decompose_type
from .type_traits import decompose_class
from .type_traits import base_type

from .type_traits import is_bool
from .type_traits import is_same
from .type_traits import is_void
from .type_traits import is_void_pointer
from .type_traits import is_const
from .type_traits import is_array
from .type_traits import is_pointer
from .type_traits import is_volatile
from .type_traits import is_integral
from .type_traits import is_reference
from .type_traits import is_arithmetic
from .type_traits import is_fundamental
from .type_traits import is_floating_point
from .type_traits import is_std_string
from .type_traits import is_std_wstring
from .type_traits import is_std_ostream
from .type_traits import is_std_wostream
from .type_traits import is_calldef_pointer
from .type_traits import is_elaborated

from .type_traits import array_size
from .type_traits import array_item_type

from .type_traits import remove_cv
from .type_traits import remove_const
from .type_traits import remove_pointer
from .type_traits import remove_volatile
from .type_traits import remove_reference
from .type_traits import remove_declarated
from .type_traits import remove_alias
from .type_traits import remove_elaborated

from .has_operator_matcher import has_public_binary_operator
from .has_operator_matcher import has_public_equal
from .has_operator_matcher import has_public_less

from .type_traits_classes import is_enum
from .type_traits_classes import enum_declaration
from .type_traits_classes import enum_traits
from .type_traits_classes import is_class
from .type_traits_classes import class_traits
from .type_traits_classes import is_class_declaration
from .type_traits_classes import class_declaration_traits
from .type_traits_classes import is_base_and_derived
from .type_traits_classes import is_convertible
from .type_traits_classes import is_noncopyable
from .type_traits_classes import is_copy_constructor
from .type_traits_classes import is_trivial_constructor
from .type_traits_classes import is_struct
from .type_traits_classes import is_union

from .type_traits_classes import is_unary_operator
from .type_traits_classes import is_binary_operator

from .type_traits_classes import has_destructor
from .type_traits_classes import has_copy_constructor

from .type_traits_classes import has_public_assign
from .type_traits_classes import has_public_destructor
from .type_traits_classes import has_public_constructor
from .type_traits_classes import has_trivial_constructor
from .type_traits_classes import find_trivial_constructor
from .type_traits_classes import find_copy_constructor
from .type_traits_classes import find_noncopyable_vars
from .type_traits_classes import has_any_non_copyconstructor
from .type_traits_classes import has_vtable

from .pointer_traits import auto_ptr_traits
from .pointer_traits import smart_pointer_traits
from .pointer_traits import internal_type_traits

from .container_traits import list_traits
from .container_traits import deque_traits
from .container_traits import queue_traits
from .container_traits import priority_queue_traits
from .container_traits import vector_traits
from .container_traits import stack_traits
from .container_traits import map_traits
from .container_traits import multimap_traits
from .container_traits import hash_map_traits
from .container_traits import hash_multimap_traits
from .container_traits import set_traits
from .container_traits import hash_set_traits
from .container_traits import multiset_traits
from .container_traits import hash_multiset_traits
from .container_traits import find_container_traits

from .container_traits import unordered_map_traits
from .container_traits import unordered_multimap_traits
from .container_traits import unordered_set_traits
from .container_traits import unordered_multiset_traits

from .function_traits import is_same_function

from . import templates
from . import call_invocation

from .dependencies import get_dependencies_from_decl
from .dependencies import dependency_info_t

from .decl_factory import decl_factory_t

from .matchers import matcher_base_t
from .matchers import or_matcher_t
from .matchers import and_matcher_t
from .matchers import not_matcher_t
from .matchers import regex_matcher_t
from .matchers import custom_matcher_t
from .matchers import access_type_matcher_t
from .matchers import virtuality_type_matcher_t

from .declarations_matchers import declaration_matcher_t
from .declarations_matchers import calldef_matcher_t
from .declarations_matchers import namespace_matcher_t
from .declarations_matchers import variable_matcher_t
from .declarations_matchers import operator_matcher_t

from .mdecl_wrapper import mdecl_wrapper_t

from .decl_printer import decl_printer_t
from .decl_printer import dump_declarations
from .decl_printer import print_declarations

from . import scopedef

from .container_traits import all_container_traits
from .container_traits import sequential_container_traits

from .runtime_errors import declaration_not_found_t
from .runtime_errors import multiple_declarations_found_t
from .runtime_errors import visit_function_has_not_been_found_t

# make matchers to look more like functors
or_matcher = or_matcher_t
"""see :class:`or_matcher_t` for documentation"""
and_matcher = and_matcher_t
"""see :class:`and_matcher_t` for documentation"""
not_matcher = not_matcher_t
"""see :class:`not_matcher_t` for documentation"""
declaration_matcher = declaration_matcher_t
"""see :class:`declaration_matcher_t` for documentation"""
calldef_matcher = calldef_matcher_t
"""see :class:`calldef_matcher_t` for documentation"""
namespace_matcher = namespace_matcher_t
"""see :class:`namespace_matcher_t` for documentation"""
variable_matcher = variable_matcher_t
"""see :class:`variable_matcher_t` for documentation"""
regex_matcher = regex_matcher_t
"""see :class:`regex_matcher_t` for documentation"""
access_type_matcher = access_type_matcher_t
"""see :class:`access_type_matcher_t` for documentation"""
operator_matcher = operator_matcher_t
"""see :class:`operator_matcher_t` for documentation"""
custom_matcher = custom_matcher_t
"""see :class:`custom_matcher_t` for documentation"""
virtuality_type_matcher = virtuality_type_matcher_t
"""see :class:`virtuality_type_matcher_t` for documentation"""

scopedef.scopedef_t._impl_all_decl_types = [
    scopedef.scopedef_t,
    enumeration_t,
    namespace_t,
    class_t,
    class_declaration_t,
    typedef_t,
    variable_t,
    calldef_t,
    member_calldef_t,
    free_calldef_t,
    operator_t,
    member_function_t,
    constructor_t,
    destructor_t,
    member_operator_t,
    casting_operator_t,
    free_function_t,
    free_operator_t]

__impl_matchers = scopedef.scopedef_t._impl_matchers
__impl_decl_types = scopedef.scopedef_t._impl_decl_types

__impl_matchers[scopedef.scopedef_t.decl] = declaration_matcher_t

__impl_matchers[scopedef.scopedef_t.class_] = declaration_matcher_t
__impl_decl_types[scopedef.scopedef_t.class_] = class_t

__impl_matchers[scopedef.scopedef_t.variable] = variable_matcher_t

__impl_matchers[scopedef.scopedef_t.calldef] = calldef_matcher_t
__impl_decl_types[scopedef.scopedef_t.calldef] = calldef_t

__impl_matchers[scopedef.scopedef_t.operator] = operator_matcher_t
__impl_decl_types[scopedef.scopedef_t.operator] = operator_t

__impl_matchers[scopedef.scopedef_t.member_function] = calldef_matcher_t
__impl_decl_types[scopedef.scopedef_t.member_function] = member_function_t

__impl_matchers[scopedef.scopedef_t.constructor] = calldef_matcher_t
__impl_decl_types[scopedef.scopedef_t.constructor] = constructor_t

__impl_matchers[scopedef.scopedef_t.member_operator] = operator_matcher_t
__impl_decl_types[scopedef.scopedef_t.member_operator] = member_operator_t

__impl_matchers[scopedef.scopedef_t.member_operator] = operator_matcher_t
__impl_decl_types[scopedef.scopedef_t.member_operator] = member_operator_t

__impl_matchers[scopedef.scopedef_t.casting_operator] = calldef_matcher_t
__impl_decl_types[scopedef.scopedef_t.casting_operator] = casting_operator_t

__impl_matchers[scopedef.scopedef_t.enumeration] = declaration_matcher_t
__impl_decl_types[scopedef.scopedef_t.enumeration] = enumeration_t

__impl_matchers[scopedef.scopedef_t.typedef] = declaration_matcher_t
__impl_decl_types[scopedef.scopedef_t.typedef] = typedef_t

__impl_matchers[namespace_t.namespace] = namespace_matcher_t

__impl_matchers[namespace_t.free_function] = calldef_matcher_t
__impl_decl_types[namespace_t.free_function] = free_function_t

__impl_matchers[namespace_t.free_operator] = operator_matcher_t
__impl_decl_types[namespace_t.free_operator] = free_operator_t



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt


class elaborated_info(object):
    """
    This class stores the name of the elaborated type specifier.

    """

    def __init__(self, elaborated_type_specifier):
        self._elaborated_type_specifier = elaborated_type_specifier

    @property
    def elaborated_type_specifier(self):
        """
        Elaborated specifier (can be: struct, union, class or enum).

        Returns:
            str: elaborated specifier
        """
        return self._elaborated_type_specifier



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines classes, that describes "callable" declarations

This modules contains definition for next C++ declarations:
    - operator
        - member
        - free
    - function
        - member
        - free
    - constructor
    - destructor
"""
from . import cpptypes
from . import declaration
from . import calldef_types


# First level in hierarchy of calldef
class argument_t(object):

    """
    class, that describes argument of "callable" declaration
    """

    def __init__(
            self,
            name='',
            decl_type=None,
            default_value=None,
            attributes=None):
        object.__init__(self)

        self._name = name
        self._default_value = default_value
        self._decl_type = decl_type
        self._attributes = attributes

    def clone(self, **keywd):
        """constructs new argument_t instance

        return argument_t(
            name=keywd.get('name', self.name),
            decl_type=keywd.get('decl_type', self.decl_type),
            default_value=keywd.get('default_value', self.default_value),
            attributes=keywd.get('attributes', self.attributes ))

        """
        return argument_t(
            name=keywd.get('name', self.name),
            decl_type=keywd.get('decl_type', self.decl_type),
            default_value=keywd.get('default_value', self.default_value),
            attributes=keywd.get('attributes', self.attributes))

    def __str__(self):
        if self.ellipsis:
            return "..."

        if self.default_value is None:
            return "%s %s" % (self.decl_type, self.name)

        return "%s %s=%s" % (self.decl_type, self.name, self.default_value)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.name == other.name \
            and self.default_value == other.default_value \
            and self.decl_type == other.decl_type

    def __hash__(self):
        return (hash(self.__class__) ^
                hash(self.name) ^
                hash(self.default_value) ^
                hash(self.decl_type))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return self.__class__.__name__ < other.__class__.__name__
        if other.default_value is None:
            return False
        if self.default_value is None:
            return self.name < other.name \
                and self.decl_type < other.decl_type
        return self.name < other.name \
            and self.default_value < other.default_value \
            and self.decl_type < other.decl_type

    @property
    def name(self):
        """Argument name.
            @type: str"""
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def ellipsis(self):
        """bool, if True argument represents ellipsis ( "..." )
        in function definition"""
        return isinstance(self.decl_type, cpptypes.ellipsis_t)

    @property
    def default_value(self):
        """Argument's default value or None.
            @type: str"""
        return self._default_value

    @default_value.setter
    def default_value(self, default_value):
        self._default_value = default_value

    @property
    def decl_type(self):
        return self._decl_type

    @decl_type.setter
    def decl_type(self, decl_type):
        self._decl_type = decl_type

    @property
    def attributes(self):
        """GCCXML attributes, set using __attribute__((gccxml("...")))
            @type: str"""
        return self._attributes

    @attributes.setter
    def attributes(self, attributes):
        self._attributes = attributes


class calldef_t(declaration.declaration_t):

    """base class for all "callable" declarations"""

    def __init__(
            self,
            name='',
            arguments=None,
            exceptions=None,
            return_type=None,
            has_extern=False,
            does_throw=True,
            mangled=None):
        declaration.declaration_t.__init__(self, name)
        if not arguments:
            arguments = []
        self._arguments = arguments
        if not exceptions:
            exceptions = []
        self._does_throw = does_throw
        self._exceptions = exceptions
        self._return_type = return_type
        self._has_extern = has_extern
        self._calling_convention = None
        self._has_inline = None
        self._mangled = mangled
        self._overrides = None

    def _get__cmp__call_items(self):
        """
        Implementation detail.

        """

        raise NotImplementedError()

    def _get__cmp__items(self):
        """
        Implementation detail.

        """

        items = [
            self.arguments,
            self.return_type,
            self.has_extern,
            self.does_throw,
            self.exceptions.sort(),
            self.has_inline]

        items.extend(self._get__cmp__call_items())
        return items

    def __eq__(self, other):
        if not declaration.declaration_t.__eq__(self, other):
            return False

        return self.return_type == other.return_type \
            and self.arguments == other.arguments \
            and self.has_extern == other.has_extern \
            and self.does_throw == other.does_throw \
            and self.exceptions.sort() == other.exceptions.sort()

    def __hash__(self):
        return (super(calldef_t, self).__hash__() ^
                hash(self.return_type) ^
                hash(self.name))

    @property
    def arguments(self):
        """The argument list.
            @type: list of :class:`argument_t`"""
        return self._arguments

    @arguments.setter
    def arguments(self, arguments):
        self._arguments = arguments

    @property
    def has_ellipsis(self):
        return self.arguments and self.arguments[-1].ellipsis

    @property
    def argument_types(self):
        """list of all argument types"""
        return [arg.decl_type for arg in self.arguments]

    @property
    def required_args(self):
        """list of all required arguments"""
        r_args = []
        for arg in self.arguments:
            if not arg.default_value:
                r_args.append(arg)
            else:
                break
        return r_args

    @property
    def optional_args(self):
        """list of all optional arguments, the arguments that have default
        value"""
        return self.arguments[len(self.required_args):]

    @property
    def overrides(self):
        """If a function is marked as an overrides, contains
        the declaration which this function overrides."""
        return self._overrides

    @overrides.setter
    def overrides(self, overrides):
        self._overrides = overrides

    @property
    def does_throw(self):
        """If False, than function does not throw any exception.
            In this case, function was declared with empty throw
            statement."""
        return self._does_throw

    @does_throw.setter
    def does_throw(self, does_throw):
        self._does_throw = does_throw

    @property
    def exceptions(self):
        """
        The list of exceptions.

        Useless with c++17 and above as dynamich exceptions
        are not allowed anymore:
        https://developers.redhat.com/articles/2021/08/06/porting-your-code-c17-gcc-11#exception_specification_changes # noqa
        
            @type: list of :class:`declaration_t`
        """
        return self._exceptions

    @exceptions.setter
    def exceptions(self, exceptions):
        """
        Useless with c++17 and above as dynamich exceptions
        are not allowed anymore:
        https://developers.redhat.com/articles/2021/08/06/porting-your-code-c17-gcc-11#exception_specification_changes # noqa
        """
        self._exceptions = exceptions

    @property
    def return_type(self):
        """The type of the return value of the "callable" or None
            (constructors).
            @type: :class:`type_t`"""
        return self._return_type

    @return_type.setter
    def return_type(self, return_type):
        self._return_type = return_type

    @property
    def overloads(self):
        """A list of overloaded "callables" (i.e. other callables with the
        same name within the same scope.

        @type: list of :class:`calldef_t`
        """
        if not self.parent:
            return []
        # finding all functions with the same name
        return self.parent.calldefs(
            name=self.name,
            function=lambda decl: decl is not self,
            allow_empty=True,
            recursive=False)

    @property
    def has_extern(self):
        """Was this callable declared as "extern"?
            @type: bool"""
        return self._has_extern

    @has_extern.setter
    def has_extern(self, has_extern):
        self._has_extern = has_extern

    @property
    def has_inline(self):
        """Was this callable declared with "inline" specifier
            @type: bool"""
        return self._has_inline

    @has_inline.setter
    def has_inline(self, has_inline):
        self._has_inline = has_inline

    def _report(self, *args, **keywd):
        # Implementation detail. Will be removed when the deprecated
        # i_depend_on_them method is dropped
        from . import dependencies  # pylint: disable=R0401
        return dependencies.dependency_info_t(self, *args, **keywd)

    def i_depend_on_them(self, recursive=True):
        self._warn_deprecated()
        answer = []
        if self.return_type:
            answer.append(
                self._report(self.return_type, hint="return type"))
        for arg in self.arguments:
            answer.append(self._report(arg.decl_type))
        for exc in self.exceptions:
            answer.append(self._report(exc, hint="exception"))
        return answer

    # pylint: disable=R0201
    def guess_calling_convention(self):
        """This function should be overriden in the derived classes and return
        more-or-less successfull guess about calling convention"""
        return calldef_types.CALLING_CONVENTION_TYPES.UNKNOWN

    @property
    def calling_convention(self):
        """function calling convention. See
            :class:CALLING_CONVENTION_TYPES class for possible values"""
        if self._calling_convention is None:
            self._calling_convention = \
                calldef_types.CALLING_CONVENTION_TYPES.extract(self.attributes)
            if not self._calling_convention:
                self._calling_convention = self.guess_calling_convention()
        return self._calling_convention

    @calling_convention.setter
    def calling_convention(self, cc):
        self._calling_convention = cc

    @property
    def mangled(self):
        """
        Unique declaration name generated by the compiler.

        :return: the mangled name
        :rtype: str

        """

        return self.get_mangled_name()

    @mangled.setter
    def mangled(self, mangled):
        self._mangled = mangled



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

"""
defines class, that describes C++ `enum`
"""

import copy
from . import declaration
from . import byte_info
from . import elaborated_info


class enumeration_t(
        declaration.declaration_t,
        byte_info.byte_info,
        elaborated_info.elaborated_info):

    """
    describes C++ `enum`
    """

    def __init__(self, name='', values=None):
        """creates class that describes C++ `enum` declaration

        The items of the list 'values' may either be strings containing
        the enumeration value name or tuples (name, numeric value).

        :param name: `enum` name
        :type name: str
        :param parent: Parent declaration
        :type parent: declaration_t
        :param values: Enumeration values
        :type values: list
        """
        declaration.declaration_t.__init__(self, name)
        byte_info.byte_info.__init__(self)
        elaborated_info.elaborated_info.__init__(self, "enum")

        # A list of tuples (valname(str), valnum(int)). The order of the list
        # should be the same as the order in the C/C++ source file.
        self._values = []

        # Initialize values via property access
        self.values = values

    def __eq__(self, other):
        if not declaration.declaration_t.__eq__(self, other):
            return False
        return self.values == other.values

    __hash__ = declaration.declaration_t.__hash__

    def _get__cmp__items(self):
        """implementation details"""
        return [self.values]

    @property
    def values(self):
        """A list of tuples (valname(str), valnum(int)) that
            contain the enumeration values.
            @type: list"""
        return copy.copy(self._values)

    @values.setter
    def values(self, values):
        self._values = []
        # None is treated like an empty list
        if values is None:
            return
        # Check that we have indeed a list...
        if not isinstance(values, list):
            raise ValueError(
                "'values' must be a list (got a %s instead)" %
                type(values).__name__)
        # Append the items individually. This has the effect that there's
        # some additional type checking and that a copy of 'values' is stored
        # and the caller cannot further manipulate the list via his own
        # reference
        for item in values:
            if isinstance(item, str):
                self.append_value(item)
            elif isinstance(item, tuple):
                name, num = item
                self.append_value(name, num)
            else:
                raise ValueError(
                    "'values' contains an invalid item: %s" % item)

    def append_value(self, valuename, valuenum=None):
        """Append another enumeration value to the `enum`.

        The numeric value may be None in which case it is automatically
        determined by increasing the value of the last item.

        When the 'values' attribute is accessed the resulting list will be in
        the same order as append_value() was called.

        :param valuename: The name of the value.
        :type valuename: str
        :param valuenum: The numeric value or None.
        :type valuenum: int
        """
        # No number given? Then use the previous one + 1
        if valuenum is None:
            if not self._values:
                valuenum = 0
            else:
                valuenum = self._values[-1][1] + 1

        # Store the new value
        self._values.append((valuename, int(valuenum)))

    def has_value_name(self, name):
        """Check if this `enum` has a particular name among its values.

        :param name: Enumeration value name
        :type name: str
        :rtype: True if there is an enumeration value with the given name
        """
        for val, _ in self._values:
            if val == name:
                return True
        return False

    def get_name2value_dict(self):
        """returns a dictionary, that maps between `enum` name( key ) and
        `enum` value( value )"""
        x = {}
        for val, num in self._values:
            x[val] = num
        return x

    def i_depend_on_them(self, recursive=True):
        self._warn_deprecated()
        return []



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt


def declaration_path(decl):
    """
    Returns a list of parent declarations names.

    Args:
        decl (declaration_t): declaration for which declaration path
                              should be calculated.

    Returns:
        list[(str | basestring)]: list of names, where first item is the top
                                  parent name and last item the inputted
                                  declaration name.
    """

    if not decl:
        return []
    if not decl.cache.declaration_path:
        result = [decl.name]
        parent = decl.parent
        while parent:
            if parent.cache.declaration_path:
                result.reverse()
                decl.cache.declaration_path = parent.cache.declaration_path + \
                    result
                return decl.cache.declaration_path
            else:
                result.append(parent.name)
                parent = parent.parent
        result.reverse()
        decl.cache.declaration_path = result
        return result

    return decl.cache.declaration_path


def partial_declaration_path(decl):
    """
    Returns a list of parent declarations names without template arguments that
    have default value.

    Args:
        decl (declaration_t): declaration for which the partial declaration
                              path should be calculated.

    Returns:
        list[(str | basestring)]: list of names, where first item is the top
                                  parent name and last item the inputted
                                  declaration name.
    """

    # TODO:
    # If parent declaration cache already has declaration_path, reuse it for
    # calculation.
    if not decl:
        return []
    if not decl.cache.partial_declaration_path:
        result = [decl.partial_name]
        parent = decl.parent
        while parent:
            if parent.cache.partial_declaration_path:
                result.reverse()
                decl.cache.partial_declaration_path \
                    = parent.cache.partial_declaration_path + result
                return decl.cache.partial_declaration_path
            else:
                result.append(parent.partial_name)
                parent = parent.parent
        result.reverse()
        decl.cache.partial_declaration_path = result
        return result

    return decl.cache.partial_declaration_path


def full_name_from_declaration_path(dpath):
    # Here I have lack of knowledge:
    # TODO: "What is the full name of declaration declared in unnamed
    # namespace?"
    result = [_f for _f in dpath if _f]
    result = result[0] + '::'.join(result[1:])
    return result


def full_name(decl, with_defaults=True):
    """
    Returns declaration full qualified name.

    If `decl` belongs to anonymous namespace or class, the function will return
    C++ illegal qualified name.

    Args:
        decl (declaration_t): declaration for which the full qualified name
                              should be calculated.

    Returns:
        list[(str | basestring)]: full name of the declaration.

    """

    if None is decl:
        raise RuntimeError("Unable to generate full name for None object!")
    if with_defaults:
        if not decl.cache.full_name:
            path = declaration_path(decl)
            if path == [""]:
                # Declarations without names are allowed (for examples class
                # or struct instances). In this case set an empty name..
                decl.cache.full_name = ""
            else:
                decl.cache.full_name = full_name_from_declaration_path(path)
        return decl.cache.full_name
    else:
        if not decl.cache.full_partial_name:
            path = partial_declaration_path(decl)
            if path == [""]:
                # Declarations without names are allowed (for examples class
                # or struct instances). In this case set an empty name.
                decl.cache.full_partial_name = ""
            else:
                decl.cache.full_partial_name = \
                    full_name_from_declaration_path(path)
        return decl.cache.full_partial_name


def get_named_parent(decl):
    """
    Returns a reference to a named parent declaration.

    Args:
        decl (declaration_t): the child declaration

    Returns:
        declaration_t: the declaration or None if not found.

    """

    if not decl:
        return None

    parent = decl.parent
    while parent and (not parent.name or parent.name == '::'):
        parent = parent.parent
    return parent



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

from .. import declarations


def bind_aliases(decls):
    """
    This function binds between class and it's typedefs.

    :param decls: list of all declarations

    :rtype: None

    """

    visited = set()
    typedefs = [
        decl for decl in decls if isinstance(decl, declarations.typedef_t)]
    for decl in typedefs:
        type_ = declarations.remove_alias(decl.decl_type)
        if not isinstance(type_, declarations.declarated_t):
            continue
        cls_inst = type_.declaration
        if not isinstance(cls_inst, declarations.class_types):
            continue
        if id(cls_inst) not in visited:
            visited.add(id(cls_inst))
            del cls_inst.aliases[:]
        cls_inst.aliases.append(decl)


def join_declarations(namespace):
    _join_namespaces(namespace)
    for ns in namespace.declarations:
        if isinstance(ns, declarations.namespace_t):
            join_declarations(ns)


def _join_namespaces(namespace):
    ddhash = {}
    decls = []

    for decl in namespace.declarations:
        _fill_declarations(ddhash, decls, decl)

    class_t = declarations.class_t
    class_declaration_t = declarations.class_declaration_t
    if class_t in ddhash and class_declaration_t in ddhash:
        # If there is a class and its forward declaration in the namespace,
        # Remove the second one from the declaration tree
        _remove_second_class(ddhash, decls, class_t, class_declaration_t)

    namespace.declarations = decls


def _fill_declarations(ddhash, decls, decl):
    if decl.__class__ not in ddhash:
        ddhash[decl.__class__] = {decl.name: [decl]}
        decls.append(decl)
    else:
        joined_decls = ddhash[decl.__class__]
        if decl.name not in joined_decls:
            decls.append(decl)
            joined_decls[decl.name] = [decl]
        else:
            if isinstance(decl, declarations.calldef_t):
                if decl not in joined_decls[decl.name]:
                    # functions has overloading
                    decls.append(decl)
                    joined_decls[decl.name].append(decl)
            elif isinstance(decl, declarations.enumeration_t):
                # unnamed enums
                if not decl.name and decl not in \
                        joined_decls[decl.name]:
                    decls.append(decl)
                    joined_decls[decl.name].append(decl)
            elif isinstance(decl, declarations.class_t):
                # unnamed classes
                if not decl.name and decl not in \
                        joined_decls[decl.name]:
                    decls.append(decl)
                    joined_decls[decl.name].append(decl)
            elif isinstance(decl, declarations.namespace_t):
                joined_decls[decl.name][0].take_parenting(decl)


def _remove_second_class(ddhash, decls, class_t, class_declaration_t):
    class_names = set()
    for name, same_name_classes in ddhash[class_t].items():
        if not name:
            continue
        if same_name_classes[0].mangled:
            class_names.add(same_name_classes[0].mangled)
        else:
            class_names.add(same_name_classes[0].name)

    class_declarations = ddhash[class_declaration_t]
    for name, same_name_class_declarations in \
            class_declarations.items():
        if not name:
            continue
        for class_declaration in same_name_class_declarations:
            if class_declaration.mangled is not None:
                # gccxml
                if class_declaration.mangled in class_names:
                    decls.remove(class_declaration)
            elif class_declaration.name in class_names:
                # castxml
                decls.remove(class_declaration)



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt

# keep py2exe happy
import xml.etree.ElementTree  # pylint: disable=W0611

import xml.etree.cElementTree as ElementTree

from . import scanner


class ietree_scanner_t(scanner.scanner_t):

    def __init__(self, xml_file, decl_factory, *args):
        scanner.scanner_t.__init__(self, xml_file, decl_factory, *args)

    def read(self):
        context = ElementTree.iterparse(
            self.xml_file,
            events=("start", "end"))
        for event, elem in context:
            if event == 'start':
                self.startElement(elem.tag, elem.attrib)
            else:
                self.endElement(elem.tag)
                elem.clear()
        self.endDocument()



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt


class byte_info(object):

    """
    This class stores information about the byte size and byte align
    values from a declaration/type.

    """

    def __init__(self):
        self._byte_size = 0
        self._byte_align = 0

    @property
    def byte_size(self):
        """
        Size of this declaration/type in bytes

        Returns:
            int: Size of this declaration/type in bytes
        """
        return self._byte_size

    @byte_size.setter
    def byte_size(self, new_byte_size):
        """
        Set size of this declaration/type in bytes

        Args:
            new_byte_size (int): Size of this declaration/type in bytes
        """
        self._byte_size = new_byte_size

    @property
    def byte_align(self):
        """
        Alignment of this declaration/type in bytes

        Returns:
            int: Alignment of this declaration/type in bytes
        """
        return self._byte_align

    @byte_align.setter
    def byte_align(self, new_byte_align):
        """
        Set size of alignment of this declaration/type in bytes

        Args:
            new_byte_align (int): Alignment of this declaration/type in bytes
        """
        self._byte_align = new_byte_align



# Copyright 2014-2017 Insight Software Consortium.
# Copyright 2004-2009 Roman Yakovenko.
# Distributed under the Boost Software License, Version 1.0.
# See http://www.boost.org/LICENSE_1_0.txt


class declaration_not_found_t(RuntimeError):
    """Exception raised when the declaration could not be found"""

    def __init__(self, decl_matcher):
        RuntimeError.__init__(self)
        self._decl_matcher = decl_matcher

    def __str__(self):
        return (
            "Unable to find declaration. Matcher: [%s]" % str(
                self._decl_matcher)
        )


class multiple_declarations_found_t(RuntimeError):
    """Exception raised when more than one declaration was found"""

    def __init__(self, decl_matcher):
        RuntimeError.__init__(self)
        self._decl_matcher = decl_matcher

    def __str__(self):
        return (
            "Multiple declarations have been found. Matcher: [%s]" % str(
                self._decl_matcher)
        )


class visit_function_has_not_been_found_t(RuntimeError):
    """
    Exception that is raised, from :func:`apply_visitor`, when a visitor could
    not be applied.

    """

    def __init__(self, visitor, decl_inst):
        RuntimeError.__init__(self)
        self.__msg = (
            "Unable to find visit function. Visitor class: %s. " +
            "Declaration instance class: %s'") \
            % (visitor.__class__.__name__, decl_inst.__class__.__name__)

    def __str__(self):
        return self.__msg
